// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"




  INLINE void
bx_cpu_c::JO_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (get_OF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jo_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JNO_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (!get_OF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jno_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JB_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (get_CF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jb_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JNB_Jb()
{
  Bit8s offset;


  offset = (Bit8s) fetch_next_byte();

  if (!get_CF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnb_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JZ_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (get_ZF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jz_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JNZ_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (!get_ZF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnz_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JBE_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (get_CF() || get_ZF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jbe_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JNBE_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (!get_CF() && !get_ZF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnbe_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JS_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (get_SF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("js_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JNS_Jb()
{
  Bit8s offset;


  offset = (Bit8s) fetch_next_byte();

  if (!get_SF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jns_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JP_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (get_PF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jp_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JNP_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (!get_PF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnp_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JL_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if ( get_SF() != get_OF() ) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jl_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JNL_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (get_SF() == get_OF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnl_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JLE_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if (get_ZF() || (get_SF() != get_OF()) ) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jle_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JNLE_Jb()
{
  Bit8s offset;

  offset = (Bit8s) fetch_next_byte();

  if ( (get_SF() == get_OF())  &&  !get_ZF()) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnle_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}

  INLINE void
bx_cpu_c::JCXZ_Jb()
{
  Bit8s offset;
  Bit32u temp_ECX;

  offset = (Bit8s) fetch_next_byte();

  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    temp_ECX = ECX;
  else
    temp_ECX = CX;

  if ( temp_ECX == 0 ) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jcxz_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
#endif
    EIP = new_EIP;
    revalidate_prefetch_q();
    }
}


  INLINE void
bx_cpu_c::RETnear_Iw()
{
  Bit16u imm16;
  Bit32u temp_ESP;

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
    temp_ESP = ESP;
  else
    temp_ESP = SP;

  imm16 = fetch_next_word();

  invalidate_prefetch_q();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u return_EIP;
    
    if (protected_mode()) {
      if ( !can_pop(4) ) {
        bx_panic("retnear_iw: can't pop EIP\n");
        /* ??? #SS(0) -or #GP(0) */
        return;
        }
      
      access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 0,
        4, CPL==3, BX_READ, &return_EIP);
  
      if (protected_mode() &&
          (return_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) ) {
        bx_panic("retnear_iw: EIP > limit\n");
        }
  
      /* Pentium book says imm16 is number of words ??? */
      if ( !can_pop(4 + imm16) ) {
        bx_panic("retnear_iw: can't release bytes from stack\n");
        /* #GP(0) -or #SS(0) ??? */
        }
  
      BX_CPU_THIS_PTR eip = return_EIP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += 4 + imm16; /* ??? should it be 2*imm16 ? */
      else
        SP  += 4 + imm16;
      }
    else {
      pop_32(&return_EIP);
      BX_CPU_THIS_PTR eip = return_EIP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += imm16; /* ??? should it be 2*imm16 ? */
      else
        SP  += imm16;
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u return_IP;

    if (protected_mode()) {
      if ( !can_pop(2) ) {
        bx_panic("retnear_iw: can't pop IP\n");
        /* ??? #SS(0) -or #GP(0) */
        return;
        }
      
      access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 0,
        2, CPL==3, BX_READ, &return_IP);
  
      if ( return_IP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("retnear_iw: IP > limit\n");
        }
  
      if ( !can_pop(2 + imm16) ) {
        bx_panic("retnear_iw: can't release bytes from stack\n");
        /* #GP(0) -or #SS(0) ??? */
        }
  
      BX_CPU_THIS_PTR eip = return_IP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += 2 + imm16; /* ??? should it be 2*imm16 ? */
      else
        SP  += 2 + imm16;
      }
    else {
      pop_16(&return_IP);
      BX_CPU_THIS_PTR eip = return_IP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += imm16; /* ??? should it be 2*imm16 ? */
      else
        SP  += imm16;
      }
    }
}

  INLINE void
bx_cpu_c::RETnear()
{
  Bit32u temp_ESP;

  invalidate_prefetch_q();

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
    temp_ESP = ESP;
  else
    temp_ESP = SP;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u return_EIP;
    
    if (protected_mode()) {
      if ( !can_pop(4) ) {
        bx_panic("retnear: can't pop EIP\n");
        /* ??? #SS(0) -or #GP(0) */
        return;
        }
    
      access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 0,
        4, CPL==3, BX_READ, &return_EIP);

      if ( return_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("retnear: EIP > limit\n");
        //exception(BX_GP_EXCEPTION, 0, 0);
        }
      BX_CPU_THIS_PTR eip = return_EIP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += 4;
      else
        SP  += 4;
      }
    else {
      pop_32(&return_EIP);
      BX_CPU_THIS_PTR eip = return_EIP;
      }
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u return_IP;

    if (protected_mode()) {
      if ( !can_pop(2) ) {
        bx_panic("retnear: can't pop IP\n");
        /* ??? #SS(0) -or #GP(0) */
        return;
        }
    
      access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 0,
        2, CPL==3, BX_READ, &return_IP);

      if ( return_IP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("retnear: IP > limit\n");
        }

      BX_CPU_THIS_PTR eip = return_IP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += 2;
      else
        SP  += 2;
      }
    else {
      pop_16(&return_IP);
      BX_CPU_THIS_PTR eip = return_IP;
      }
    }
}

  INLINE void
bx_cpu_c::RETfar_Iw()
{
  Bit16s imm16;

  /* ??? is imm16, number of bytes/words depending on operandsize ? */

  imm16 = fetch_next_word();

  invalidate_prefetch_q();

#if BX_CPU >= 2
  if (protected_mode()) {
    if (bx_dbg._protected) bx_printf("RETfar_Iw: entering\n");
    BX_CPU_THIS_PTR return_protected(imm16);
    return;
    }
#endif

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eip, ecs_raw;

    pop_32(&eip);
    pop_32(&ecs_raw);
    BX_CPU_THIS_PTR eip = eip;
    load_seg_reg(&BX_CPU_THIS_PTR cs, (Bit16u) ecs_raw);
    if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
      ESP += imm16;
    else
      SP  += imm16;
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u ip, cs_raw;

    pop_16(&ip);
    pop_16(&cs_raw);
    BX_CPU_THIS_PTR eip = (Bit32u) ip;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
      ESP += imm16;
    else
      SP  += imm16;
    }
}

  INLINE void
bx_cpu_c::RETfar()
{
  invalidate_prefetch_q();

#if BX_CPU >= 2
  if ( protected_mode() ) {
    if (bx_dbg._protected) bx_printf("RETfar: entering\n");
    BX_CPU_THIS_PTR return_protected(0);
    return;
    }
#endif

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eip, ecs_raw;

    pop_32(&eip);
    pop_32(&ecs_raw); /* 32bit pop, MSW discarded */
    BX_CPU_THIS_PTR eip = eip;
    load_seg_reg(&BX_CPU_THIS_PTR cs, (Bit16u) ecs_raw);
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u ip, cs_raw;

    pop_16(&ip);
    pop_16(&cs_raw);
    BX_CPU_THIS_PTR eip = (Bit32u) ip;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
}

  INLINE void
bx_cpu_c::ENTER_IwIb()
{
#if BX_CPU < 2
  bx_panic("ENTER_IwIb: not supported by 8086!\n");
#else
  Bit16u dynamic_size;
  Bit8u  level;
  Bit32u frame_ptr32;
  Bit16u frame_ptr16;

  dynamic_size = fetch_next_word();

  level = fetch_next_byte();

  invalidate_prefetch_q();

  level %= 32;
/* ??? */
if (level) bx_panic("enter(): level > 0\n");
//if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b && BX_CPU_THIS_PTR is_32bit_opsize==0) {
//  bx_printf("enter(): stacksize!=opsize: I'm unsure of the code for this\n");
//  bx_panic("         The Intel manuals are a mess on this one!\n");
//  }

  if ( protected_mode() ) {
    Bit32u bytes_to_push, temp_ESP;

    if (level == 0) {
      if (BX_CPU_THIS_PTR is_32bit_opsize)
        bytes_to_push = 4 + dynamic_size;
      else
        bytes_to_push = 2 + dynamic_size;
      }
    else { /* level > 0 */
      if (BX_CPU_THIS_PTR is_32bit_opsize)
        bytes_to_push = 4 + (level-1)*4 + 4 + dynamic_size;
      else
        bytes_to_push = 2 + (level-1)*2 + 2 + dynamic_size;
      }
    if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
      temp_ESP = ESP;
    else
      temp_ESP = SP;
    if ( !can_push(&BX_CPU_THIS_PTR ss.cache, temp_ESP, bytes_to_push) ) {
      bx_panic("ENTER: not enough room on stack!\n");
      exception(BX_SS_EXCEPTION, 0, 0);
      return;
      }
    if (bx_dbg._protected) bx_printf("ENTER(protected mode): succeeding\n");
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize)
    push_32(EBP);
  else
    push_16(BP);

  // can just do frame_ptr32 = ESP for either case ???
  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    frame_ptr32 = ESP;
  else
    frame_ptr32 = SP;

  if (level > 0) {
    /* do level-1 times */
    while (--level) {
      if (BX_CPU_THIS_PTR is_32bit_opsize) {
        Bit32u temp32;

        if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
          EBP -= 4;
          read_virtual_dword(&BX_CPU_THIS_PTR ss, EBP, &temp32);
          ESP -= 4;
          write_virtual_dword(&BX_CPU_THIS_PTR ss, ESP, &temp32);
          }
        else { /* 16bit stacksize */
          BP -= 4;
          read_virtual_dword(&BX_CPU_THIS_PTR ss, BP, &temp32);
          SP -= 4;
          write_virtual_dword(&BX_CPU_THIS_PTR ss, SP, &temp32);
          }
        }
      else { /* 16bit opsize */
        Bit16u temp16;

        if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
          EBP -= 2;
          read_virtual_word(&BX_CPU_THIS_PTR ss, EBP, &temp16);
          ESP -= 2;
          write_virtual_word(&BX_CPU_THIS_PTR ss, ESP, &temp16);
          }
        else { /* 16bit stacksize */
          BP -= 2;
          read_virtual_word(&BX_CPU_THIS_PTR ss, BP, &temp16);
          SP -= 2;
          write_virtual_word(&BX_CPU_THIS_PTR ss, SP, &temp16);
          }
        }
      } /* while (--level) */

    /* push(frame pointer) */
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
        ESP -= 4;
        write_virtual_dword(&BX_CPU_THIS_PTR ss, ESP, &frame_ptr32);
        }
      else {
        SP -= 4;
        write_virtual_dword(&BX_CPU_THIS_PTR ss, SP, &frame_ptr32);
        }
      }
    else { /* 16bit opsize */
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
        frame_ptr16 = frame_ptr32;
        ESP -= 2;
        write_virtual_word(&BX_CPU_THIS_PTR ss, ESP, &frame_ptr16);
        }
      else {
        frame_ptr16 = frame_ptr32;
        SP -= 2;
        write_virtual_word(&BX_CPU_THIS_PTR ss, SP, &frame_ptr16);
        }
      }
    } /* if (level > 0) ... */

  if (BX_CPU_THIS_PTR is_32bit_opsize)
    EBP = frame_ptr32;
  else
    BP = frame_ptr32;

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
    ESP = ESP - dynamic_size;
    }
  else { /* 16bit stack */
    SP = SP - dynamic_size;
    }
#endif
}

  INLINE void
bx_cpu_c::LEAVE()
{
#if BX_CPU < 2
  bx_panic("LEAVE: not supported by 8086!\n");
#else
  Bit32u temp_EBP;


  invalidate_prefetch_q();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    temp_EBP = EBP;
  else
#endif
    temp_EBP = BP;

  if ( protected_mode() ) {
    if (BX_CPU_THIS_PTR ss.cache.u.segment.c_ed) { /* expand up */
      if (temp_EBP <= BX_CPU_THIS_PTR ss.cache.u.segment.limit_scaled) {
        bx_panic("LEAVE: BP > BX_CPU_THIS_PTR ss.limit\n");
        exception(BX_SS_EXCEPTION, 0, 0);
        return;
        }
      }
    else { /* normal */
      if (temp_EBP > BX_CPU_THIS_PTR ss.cache.u.segment.limit_scaled) {
        bx_panic("LEAVE: BP > BX_CPU_THIS_PTR ss.limit\n");
        exception(BX_SS_EXCEPTION, 0, 0);
        return;
        }
      }
    if (bx_dbg._protected) bx_printf("LEAVE(protected mode): succeeding\n");
    }


  // delete frame
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    ESP = EBP;
  else
#endif
    SP = BP;

  // restore frame pointer
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u temp32;

    pop_32(&temp32);
    EBP = temp32;
    }
  else
#endif
    {
    Bit16u temp16;

    pop_16(&temp16);
    BP = temp16;
    }
#endif
}

  INLINE void
bx_cpu_c::LOOPNE_Jb()
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = fetch_next_byte();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    count = BX_CPU_THIS_PTR ecx;
  else
#endif /* BX_CPU >= 3 */
    count = CX;

  count--;
  if ( (count!=0) && (get_ZF()==0) ) {

    new_EIP = EIP + imm8;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loopne_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    revalidate_prefetch_q();
    }

  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    BX_CPU_THIS_PTR ecx--;
  else
    CX--;
}

  INLINE void
bx_cpu_c::LOOPE_Jb()
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = fetch_next_byte();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    count = BX_CPU_THIS_PTR ecx;
  else
#endif /* BX_CPU >= 3 */
    count = CX;

  count--;
  if ( (count!=0) && get_ZF()) {

    new_EIP = EIP + imm8;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loope_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    revalidate_prefetch_q();
    }
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    BX_CPU_THIS_PTR ecx--;
  else
    CX--;
}

  INLINE void
bx_cpu_c::LOOP_Jb()
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = fetch_next_byte();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    count = BX_CPU_THIS_PTR ecx;
  else
#endif /* BX_CPU >= 3 */
    count = CX;

  count--;
  if (count != 0) {

    new_EIP = EIP + imm8;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loop_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    revalidate_prefetch_q();
    }
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    BX_CPU_THIS_PTR ecx--;
  else
    CX--;
}

  INLINE void
bx_cpu_c::CALL_Av()
{
  Bit32u new_EIP;


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32s disp32;

    disp32 = fetch_next_dword();
    invalidate_prefetch_q();

    new_EIP = EIP + disp32;

    if ( protected_mode() ) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("call_av: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }

    /* push 32 bit EA of next instruction */
    push_32(BX_CPU_THIS_PTR eip);
    BX_CPU_THIS_PTR eip = new_EIP;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit opsize mode */
    Bit16s disp16;


    disp16 = fetch_next_word();
    invalidate_prefetch_q();

    new_EIP = EIP + disp16;
    new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if ( protected_mode() &&
         (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) ) {
      bx_panic("call_av: new_IP > BX_CPU_THIS_PTR cs.limit\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
#endif

    /* push 16 bit EA of next instruction */
    push_16(IP);

    BX_CPU_THIS_PTR eip = new_EIP;
    }
}

  INLINE void
bx_cpu_c::CALL_Ap()
{
  Bit16u cs_raw;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u disp32;

    disp32 = fetch_next_dword();
    cs_raw     = fetch_next_word();
    invalidate_prefetch_q();

    if (protected_mode()) {
      BX_CPU_THIS_PTR call_protected(cs_raw, disp32);
      return;
      }
    push_32(BX_CPU_THIS_PTR cs.selector.value);
    push_32(BX_CPU_THIS_PTR eip);
    BX_CPU_THIS_PTR eip = disp32;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit opsize mode */
    Bit16u disp16;

    disp16 = fetch_next_word();
    cs_raw     = fetch_next_word();
    invalidate_prefetch_q();

#if BX_CPU >= 2
    if (protected_mode()) {
      BX_CPU_THIS_PTR call_protected(cs_raw, disp16);
      return;
      }
#endif
    push_16(BX_CPU_THIS_PTR cs.selector.value);
    push_16((Bit16u) BX_CPU_THIS_PTR eip);
    BX_CPU_THIS_PTR eip = (Bit32u) disp16;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
}

  INLINE void
bx_cpu_c::CALL_Ev()
{
  Bit32u temp_ESP;

  decode_exgx();

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    temp_ESP = ESP;
  else
    temp_ESP = SP;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }
    invalidate_prefetch_q();

    if (protected_mode()) {
      if (op1_32 > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("call_ev: EIP out of CS limits!\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      if ( !can_push(&BX_CPU_THIS_PTR ss.cache, temp_ESP, 4) ) {
        bx_panic("call_ev: can't push EIP\n");
        }
      }

    push_32(BX_CPU_THIS_PTR eip);

    BX_CPU_THIS_PTR eip = op1_32;
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }
    invalidate_prefetch_q();

#if BX_CPU >= 2
    if (protected_mode()) {
      if (op1_16 > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("call_ev: IP out of CS limits!\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      if ( !can_push(&BX_CPU_THIS_PTR ss.cache, temp_ESP, 2) ) {
        bx_panic("call_ev: can't push IP\n");
        }
      }
#endif

    push_16(IP);

    BX_CPU_THIS_PTR eip = op1_16;
    }
}

  INLINE void
bx_cpu_c::CALL_Ep()
{
  Bit16u cs_raw;

  decode_exgx();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      bx_panic("CALL_Ep: op1 is a register\n");
      return;
      }

    /* pointer, segment address pair */
    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+4, &cs_raw);
    invalidate_prefetch_q();

    if ( protected_mode() ) {
      BX_CPU_THIS_PTR call_protected(cs_raw, op1_32);
      return;
      }

    push_32(BX_CPU_THIS_PTR cs.selector.value);
    push_32(BX_CPU_THIS_PTR eip);

    BX_CPU_THIS_PTR eip = op1_32;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      bx_panic("CALL_Ep: op1 is a register");
      return;
      }

    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+2, &cs_raw);
    invalidate_prefetch_q();

    if ( protected_mode() ) {
      BX_CPU_THIS_PTR call_protected(cs_raw, op1_16);
      return;
      }

    push_16(BX_CPU_THIS_PTR cs.selector.value);
    push_16(IP);

    BX_CPU_THIS_PTR eip = op1_16;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
}


  INLINE void
bx_cpu_c::JMP_Jv()
{
  Bit32u new_EIP;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32s imm32;

    imm32 = fetch_next_dword();
    invalidate_prefetch_q();

    new_EIP = EIP + imm32;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit operand size mode */
    Bit16s imm16;

    imm16 = fetch_next_word();
    invalidate_prefetch_q();

    new_EIP = EIP + imm16;
    new_EIP &= 0x0000ffff;
    }

#if BX_CPU >= 2
  if (protected_mode()) {
    if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
      bx_panic("jmp_jv: offset outside of CS limits\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }
#endif

  BX_CPU_THIS_PTR eip = new_EIP;
}

  INLINE void
bx_cpu_c::JMP_Ap()
{
  Bit16u cs_raw;
  Bit32u disp32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    disp32 = fetch_next_dword();
    }
  else { /* 16bit operands */
    disp32 = fetch_next_word(); /* clear upper 16 bits */
    }

  cs_raw = fetch_next_word();
  invalidate_prefetch_q();

#if BX_CPU >= 2
  if (protected_mode()) {
    if (bx_dbg._protected) bx_printf("jmp_ap():\n");
    BX_CPU_THIS_PTR jump_protected(cs_raw, disp32);
    return;
    }
#endif

  load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
  BX_CPU_THIS_PTR eip = disp32;
}



  INLINE void
bx_cpu_c::JMP_Jb()
{
  Bit8s imm8;
  Bit32u new_EIP;

  imm8 = fetch_next_byte();
  invalidate_prefetch_q();

#if BX_WEIRDISMS
  if (imm8 == -2) {
    if (BX_CPU_THIS_PTR eflags.if_ == 0)
      bx_panic("endless loop detected!\n");
    }
#endif

  new_EIP = EIP + imm8;
  if (BX_CPU_THIS_PTR is_32bit_opsize==0)
    new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
  if (protected_mode()) {
    if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
      bx_panic("jmp_jb: offset outside of CS limits\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }
#endif

  EIP = new_EIP;
}

  INLINE void
bx_cpu_c::JO_Jv()
{
#if BX_CPU < 3
  bx_panic("JO_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (get_OF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jo_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JNO_Jv()
{
#if BX_CPU < 3
  bx_panic("JNO_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (!get_OF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jno_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JB_Jv()
{
#if BX_CPU < 3
  bx_panic("JB_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (get_CF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jb_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JNB_Jv()
{
#if BX_CPU < 3
  bx_panic("JNB_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (!get_CF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnb_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JZ_Jv()
{
#if BX_CPU < 3
  bx_panic("JZ_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (get_ZF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jz_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JNZ_Jv()
{
#if BX_CPU < 3
  bx_panic("JNZ_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (!get_ZF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnz_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JBE_Jv()
{
#if BX_CPU < 3
  bx_panic("JBE_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (get_CF() || get_ZF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jbe_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JNBE_Jv()
{
#if BX_CPU < 3
  bx_panic("JNBE_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (!get_CF() && !get_ZF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnbe_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JS_Jv()
{
#if BX_CPU < 3
  bx_panic("JS_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (get_SF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("js_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JNS_Jv()
{
#if BX_CPU < 3
  bx_panic("JNS_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (!get_SF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jns_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JP_Jv()
{
#if BX_CPU < 3
  bx_panic("JP_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (get_PF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jp_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JNP_Jv()
{
#if BX_CPU < 3
  bx_panic("JNP_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (!get_PF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnp_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JL_Jv()
{
#if BX_CPU < 3
  bx_panic("JL_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if ( get_SF() != get_OF() ) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jl_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JNL_Jv()
{
#if BX_CPU < 3
  bx_panic("JNL_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (get_SF() == get_OF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnl_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JLE_Jv()
{
#if BX_CPU < 3
  bx_panic("JLE_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if (get_ZF() || (get_SF() != get_OF()) ) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jle_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}

  INLINE void
bx_cpu_c::JNLE_Jv()
{
#if BX_CPU < 3
  bx_panic("JNLE_Jv: not supported on < 386\n");
#else
  Bit32s offset32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    offset32 = (Bit32s) fetch_next_dword();
    }
  else {
    offset32 = (Bit16s) fetch_next_word();
    }

  if ( (get_SF() == get_OF())  &&  !get_ZF()) {
    Bit32u new_EIP;

    invalidate_prefetch_q();
    new_EIP = EIP + offset32;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jnle_jv: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        return;
        }
      }
    EIP = new_EIP;
    }
#endif
}



  INLINE void
bx_cpu_c::JMP_Ev()
{
  Bit32u new_EIP;

  decode_exgx();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    invalidate_prefetch_q();
    new_EIP = op1_32;
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    invalidate_prefetch_q();
    new_EIP = op1_16;
    }

#if BX_CPU >= 2
  if (protected_mode()) {
    if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
      bx_panic("jmp_ev: IP out of CS limits!\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }
#endif

  BX_CPU_THIS_PTR eip = new_EIP;
}

  /* Far indirect jump */

  INLINE void
bx_cpu_c::JMP_Ep()
{
  Bit16u cs_raw;

  decode_exgx();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      /* far indirect must specify a memory address */
      bx_panic("JMP_Ep(): op1 is a register\n");
      return;
      }

    /* pointer, segment address pair */
    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+4, &cs_raw);
    invalidate_prefetch_q();

    if ( protected_mode() ) {
      if (bx_dbg._protected) bx_printf("jmp_ep():\n");
      BX_CPU_THIS_PTR jump_protected(cs_raw, op1_32);
      return;
      }

    BX_CPU_THIS_PTR eip = op1_32;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      /* far indirect must specify a memory address */
      bx_panic("JMP_Ep(): op1 is a register\n");
      return;
      }

    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+2, &cs_raw);
    invalidate_prefetch_q();

#if BX_CPU >= 2
    if ( protected_mode() ) {
      if (bx_dbg._protected) bx_printf("jmp_ep():\n");
      BX_CPU_THIS_PTR jump_protected(cs_raw, op1_16);
      return;
      }
#endif

    BX_CPU_THIS_PTR eip = op1_16;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
}

  INLINE void
bx_cpu_c::IRET()
{
  invalidate_prefetch_q();

  if (v8086_mode()) {
    // IOPL check in stack_return_from_v86()
    BX_CPU_THIS_PTR stack_return_from_v86();
    return;
    }

#if BX_CPU >= 2
  if (protected_mode()) {
    BX_CPU_THIS_PTR iret_protected();
    return;
    }
#endif

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eip, ecs_raw, eflags;

    pop_32(&eip);
    pop_32(&ecs_raw);
    pop_32(&eflags);

    load_seg_reg(&BX_CPU_THIS_PTR cs, (Bit16u) ecs_raw);
    BX_CPU_THIS_PTR eip = eip;
    write_eflags(eflags, 1, 1, 0, 1);
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u ip, cs_raw, flags;

    pop_16(&ip);
    pop_16(&cs_raw);
    pop_16(&flags);

    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    BX_CPU_THIS_PTR eip = (Bit32u) ip;
    write_flags(flags, 1, 1);
    }
}
