/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.

BOCHS is commercial software.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/





extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/time.h>
}


#include "bochs.h"

#if BX_SHOW_IPS
unsigned long ips_count=0;
#endif



/* end prototypes */



/* This array defines a look-up table for the even parity-ness
   of an 8bit quantity, for optimal assignment of the parity bit
   in the EFLAGS register
  */
const Boolean bx_parity_lookup[256] = {
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1
  };






#if USE_INLINE
#include "arith.c"
#include "arith2.c"
#include "bit.c"
#include "ctrl_xfer.c"
#include "data_xfer.c"
#include "flag_ctrl.c"
#include "io.c"
#include "logical.c"
#include "shift.c"
#include "ndp.c"
#include "proc_ctrl.c"
#include "protect_ctrl.c"
#include "segment_ctrl.c"
#include "soft_int.c"
#include "stack.c"
#include "string.c"
#include "bcd.c"
#endif



  void
bx_cpu_c::cpu_loop(void)
{

  Bit8u next_byte, mod_rm_byte;
  Bit8u mod, opcode, rm;

#if BX_PROVIDE_DEVICE_MODELS==0
  void bx_instruction_tick(void);
#endif


  BX_CPU_THIS_PTR errno = 0;

  (void) setjmp( BX_CPU_THIS_PTR jmp_buf_env );

  BX_CPU_LOOP_ENTRY();

  while (1) {

#if BX_CPU >= 2
    BX_CPU_THIS_PTR EXT = 0;
    BX_CPU_THIS_PTR prev_eip = BX_CPU_THIS_PTR eip;   // save instruction pointer
    BX_CPU_THIS_PTR prev_esp = BX_CPU_THIS_PTR esp;   // save stack pointer
#endif

    BX_LOOP_ITERATION();

    if (BX_CPU_THIS_PTR async_event) {
      /* INPUT PROCESSING ORDER:
       * 1) instruction exception
       * 2) single step
       * 3) NMI
       * 4) processor extension segment overrun
       * 5) INTR
       */
      /* ??? look at priorities for various cpu levels */

      /* single step recognized here */
      if ( BX_CPU_THIS_PTR single_step_event ) {
        bx_panic("decode: single step recognized\n");
        /* should BX_CPU_THIS_PTR EXT be set to 1 here ??? */
        exception(1, 0, 0); /* single step interrupt */
        BX_CPU_THIS_PTR single_step_event = 0;
        /* fall thru to other interrupt processing */
        }

      /* single step generated here */
      if ( BX_CPU_THIS_PTR eflags.tf ) {
        BX_CPU_THIS_PTR single_step_event = 1;
        bx_panic("decode: single step generated\n");
        }

      /* if last instruction caused processing of interrupts to pend,
       * such as STI or load of SS, then skip for one instruction */
      if (BX_CPU_THIS_PTR inhibit_interrupts) {
        BX_CPU_THIS_PTR inhibit_interrupts = 0;
        }

      else if ( BX_CPU_THIS_PTR INTR && BX_CPU_THIS_PTR eflags.if_ ) {
        Bit8u vector;

        vector = bx_pc_system.IAC(); /* may set INTR with next interrupt */
        //if (bx_dbg.interrupts) bx_printf("decode: interrupt %u\n",
        //                                   (unsigned) vector);
        BX_CPU_THIS_PTR errno = 0;
        BX_CPU_THIS_PTR EXT   = 1; /* external event */
        BX_CPU_THIS_PTR interrupt(vector, 0, 0, 0);
        }

      // DMA Hold Request
      else if ( bx_pc_system.HRQ ) {
        // assert Hold Acknowledge (HLDA) and go into a bus hold state
        bx_pc_system.raise_HLDA();
        }

      if ( !(BX_CPU_THIS_PTR INTR || BX_CPU_THIS_PTR fd_event ||
             BX_CPU_THIS_PTR single_step_event || bx_pc_system.HRQ) )
        BX_CPU_THIS_PTR async_event = 0;

      }

#if BX_CPU >= 3
    BX_CPU_THIS_PTR is_32bit_opsize   = BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
    BX_CPU_THIS_PTR is_32bit_addrsize = BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
#endif

    BX_CPU_THIS_PTR segment_override = NULL;
    BX_CPU_THIS_PTR rep_prefix = 0;

#if BX_CPU >= 2
    BX_CPU_THIS_PTR errno = 0;
#endif



#if 0
  {
  void bx_save_cpu_snapshot(bx_cpu_c *cpu);

  bx_save_cpu_snapshot(&bx_cpu);
  }
#endif

    /* Call debug to display current instruction.
       by conditional compile */
#if 0
#if BX_DEBUGGER
  debug(&cs, BX_CPU_THIS_PTR eip);
#endif
#endif

//if (bx_dbg.debugger) debug(& BX_CPU_THIS_PTR cs, BX_CPU_THIS_PTR eip);

//debug(& BX_CPU_THIS_PTR cs, BX_CPU_THIS_PTR eip);
//if ( (bx_cpu.cs.selector.value == 0x0767) &&
//     (bx_cpu.eip == 0x00004eb0) &&
//     (bx_cpu.es.selector.value == 0x0000) ) {
//  fprintf(stderr, "================= found CS:IP @ 0767:4eb0 ===========\n");
//  bx_panic("================= found CS:IP @ 0767:4eb0 ===========\n");
//  //debug(& BX_CPU_THIS_PTR cs, BX_CPU_THIS_PTR eip);
//  }

    // prefetch 8-byte aligned memory block into prefetch queue
    //prefetch();



    BX_ICOUNT();
#if BX_SHOW_IPS
    ips_count++;
#endif

    /* On the 386 and above, instructions must be a maximum of 15 bytes long.
     * this means redundant prefix codes can put the byte count over 15 and
     * cause an illegal instruction.
     */
    // ?? should check for >15 byte instructions
    while (1) {
      next_byte = fetch_next_byte(); 

      switch (next_byte) {

        case 0x00: BX_CPU_THIS_PTR ADD_EbGb(); break;
        case 0x01: BX_CPU_THIS_PTR ADD_EvGv(); break;
        case 0x02: BX_CPU_THIS_PTR ADD_GbEb(); break;
        case 0x03: BX_CPU_THIS_PTR ADD_GvEv(); break;
        case 0x04: BX_CPU_THIS_PTR ADD_ALIb(); break;
        case 0x05: BX_CPU_THIS_PTR ADD_eAXIv(); break;
        case 0x06: BX_CPU_THIS_PTR PUSH_ES(); break;
        case 0x07: BX_CPU_THIS_PTR POP_ES(); break;
        case 0x08: BX_CPU_THIS_PTR OR_EbGb(); break;
        case 0x09: BX_CPU_THIS_PTR OR_EvGv(); break;
        case 0x0A: BX_CPU_THIS_PTR OR_GbEb(); break;
        case 0x0B: BX_CPU_THIS_PTR OR_GvEv(); break;
        case 0x0C: BX_CPU_THIS_PTR OR_ALIb(); break;
        case 0x0D: BX_CPU_THIS_PTR OR_eAXIv(); break;
        case 0x0E: BX_CPU_THIS_PTR PUSH_CS(); break;
        case 0x0F: /* 2-byte escape */
          next_byte = fetch_next_byte(); 
          switch (next_byte) {
            case 0x00: /* Group 6 */
              mod_rm_byte = peek_next_byte();
              BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
              switch (opcode) {
                case 0x00: BX_CPU_THIS_PTR SLDT_Ew(); break;
                case 0x01: BX_CPU_THIS_PTR STR_Ew(); break;
                case 0x02: BX_CPU_THIS_PTR LLDT_Ew(); break;
                case 0x03: BX_CPU_THIS_PTR LTR_Ew(); break;
                case 0x04: BX_CPU_THIS_PTR VERR_Ew(); break;
                case 0x05: BX_CPU_THIS_PTR VERW_Ew(); break;
                case 0x06: bx_panic("decode: : Grp6:\n"); break;
                case 0x07: bx_panic("decode: : Grp6:\n"); break;
                default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
                }
              break;
            case 0x01: /* Group 7 */
              mod_rm_byte = peek_next_byte();
              BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
              switch (opcode) {
                case 0x00: BX_CPU_THIS_PTR SGDT_Ms(); break;
                case 0x01: BX_CPU_THIS_PTR SIDT_Ms(); break;
                case 0x02: BX_CPU_THIS_PTR LGDT_Ms(); break;
                case 0x03: BX_CPU_THIS_PTR LIDT_Ms(); break;
                case 0x04: BX_CPU_THIS_PTR SMSW_Ew(); break;
                case 0x05: bx_panic("decode: : Grp7:\n"); break;
                case 0x06: BX_CPU_THIS_PTR LMSW_Ew(); break;
                case 0x07: bx_panic("decode: : Grp7:\n"); break;
                default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
                }
              break;

            case 0x02: BX_CPU_THIS_PTR LAR_GvEw(); break;
            case 0x03: BX_CPU_THIS_PTR LSL_GvEw(); break;
            case 0x04: bx_panic("decode: : 0F 04:\n"); break;
            case 0x05: BX_CPU_THIS_PTR LOADALL(); break;
            case 0x06: BX_CPU_THIS_PTR CLTS(); break;
            case 0x07: bx_panic("decode: : 0F 07:\n"); break;
            case 0x08: BX_CPU_THIS_PTR INVD(); break;
            case 0x09: BX_CPU_THIS_PTR WBINVD(); break;
            case 0x0A:
            case 0x0B:
            case 0x0C:
            case 0x0D:
            case 0x0E:
            case 0x0F: bx_panic("decode: : 0F XX:\n"); break;

            case 0x10:
            case 0x11:
            case 0x12:
            case 0x13:
            case 0x14:
            case 0x15:
            case 0x16:
            case 0x17:
            case 0x18:
            case 0x19:
            case 0x1A:
            case 0x1B:
            case 0x1C:
            case 0x1D:
            case 0x1E:
            case 0x1F:
              bx_panic("decode: : 0F XX:\n"); break;
              break;

            case 0x20: BX_CPU_THIS_PTR MOV_RdCd(); break;
            case 0x21: BX_CPU_THIS_PTR MOV_RdDd(); break;
            case 0x22: BX_CPU_THIS_PTR MOV_CdRd(); break;
            case 0x23: BX_CPU_THIS_PTR MOV_DdRd(); break;
            case 0x24: BX_CPU_THIS_PTR MOV_RdTd(); break;
            case 0x25: bx_panic("decode: : 0F XX:\n"); break;
            case 0x26: BX_CPU_THIS_PTR MOV_TdRd(); break;
            case 0x27:
            case 0x28:
            case 0x29:
            case 0x2A:
            case 0x2B:
            case 0x2C:
            case 0x2D:
            case 0x2E:
            case 0x2F: bx_panic("decode: : 0F XX:\n"); break;

            case 0x30:
            case 0x31:
            case 0x32:
            case 0x33:
            case 0x34:
            case 0x35:
            case 0x36:
            case 0x37:
            case 0x38:
            case 0x39:
            case 0x3A:
            case 0x3B:
            case 0x3C:
            case 0x3D:
            case 0x3E:
            case 0x3F: bx_panic("decode: : 0F XX:\n"); break;

            case 0x40:
            case 0x41:
            case 0x42:
            case 0x43:
            case 0x44:
            case 0x45:
            case 0x46:
            case 0x47:
            case 0x48:
            case 0x49:
            case 0x4A:
            case 0x4B:
            case 0x4C:
            case 0x4D:
            case 0x4E:
            case 0x4F: bx_panic("decode: : 0F XX:\n"); break;

            case 0x50:
            case 0x51:
            case 0x52:
            case 0x53:
            case 0x54:
            case 0x55:
            case 0x56:
            case 0x57:
            case 0x58:
            case 0x59:
            case 0x5A:
            case 0x5B:
            case 0x5C:
            case 0x5D:
            case 0x5E:
            case 0x5F: bx_panic("decode: : 0F XX:\n"); break;

            case 0x60:
            case 0x61:
            case 0x62:
            case 0x63:
            case 0x64:
            case 0x65:
            case 0x66:
            case 0x67:
            case 0x68:
            case 0x69:
            case 0x6A:
            case 0x6B:
            case 0x6C:
            case 0x6D:
            case 0x6E:
            case 0x6F: bx_panic("decode: : 0F XX:\n"); break;

            case 0x70:
            case 0x71:
            case 0x72:
            case 0x73:
            case 0x74:
            case 0x75:
            case 0x76:
            case 0x77:
            case 0x78:
            case 0x79:
            case 0x7A:
            case 0x7B:
            case 0x7C:
            case 0x7D:
            case 0x7E:
            case 0x7F: bx_panic("decode: : 0F XX:\n"); break;

            case 0x80: BX_CPU_THIS_PTR JO_Jv(); break;
            case 0x81: BX_CPU_THIS_PTR JNO_Jv(); break;
            case 0x82: BX_CPU_THIS_PTR JB_Jv(); break;
            case 0x83: BX_CPU_THIS_PTR JNB_Jv(); break;
            case 0x84: BX_CPU_THIS_PTR JZ_Jv(); break;
            case 0x85: BX_CPU_THIS_PTR JNZ_Jv(); break;
            case 0x86: BX_CPU_THIS_PTR JBE_Jv(); break;
            case 0x87: BX_CPU_THIS_PTR JNBE_Jv(); break;
            case 0x88: BX_CPU_THIS_PTR JS_Jv(); break;
            case 0x89: BX_CPU_THIS_PTR JNS_Jv(); break;
            case 0x8A: BX_CPU_THIS_PTR JP_Jv(); break;
            case 0x8B: BX_CPU_THIS_PTR JNP_Jv(); break;
            case 0x8C: BX_CPU_THIS_PTR JL_Jv(); break;
            case 0x8D: BX_CPU_THIS_PTR JNL_Jv(); break;
            case 0x8E: BX_CPU_THIS_PTR JLE_Jv(); break;
            case 0x8F: BX_CPU_THIS_PTR JNLE_Jv(); break;

            case 0x90: BX_CPU_THIS_PTR SETO_Eb(); break;
            case 0x91: BX_CPU_THIS_PTR SETNO_Eb(); break;
            case 0x92: BX_CPU_THIS_PTR SETB_Eb(); break;
            case 0x93: BX_CPU_THIS_PTR SETNB_Eb(); break;
            case 0x94: BX_CPU_THIS_PTR SETZ_Eb(); break;
            case 0x95: BX_CPU_THIS_PTR SETNZ_Eb(); break;
            case 0x96: BX_CPU_THIS_PTR SETBE_Eb(); break;
            case 0x97: BX_CPU_THIS_PTR SETNBE_Eb(); break;
            case 0x98: BX_CPU_THIS_PTR SETS_Eb(); break;
            case 0x99: BX_CPU_THIS_PTR SETNS_Eb(); break;
            case 0x9A: BX_CPU_THIS_PTR SETP_Eb(); break;
            case 0x9B: BX_CPU_THIS_PTR SETNP_Eb(); break;
            case 0x9C: BX_CPU_THIS_PTR SETL_Eb(); break;
            case 0x9D: BX_CPU_THIS_PTR SETNL_Eb(); break;
            case 0x9E: BX_CPU_THIS_PTR SETLE_Eb(); break;
            case 0x9F: BX_CPU_THIS_PTR SETNLE_Eb(); break;

            case 0xA0: BX_CPU_THIS_PTR PUSH_FS(); break;
            case 0xA1: BX_CPU_THIS_PTR POP_FS(); break;
            case 0xA2: BX_CPU_THIS_PTR CPUID(); break;
            case 0xA3: BX_CPU_THIS_PTR BT_EvGv(); break;
            case 0xA4: BX_CPU_THIS_PTR SHLD_EvGvIb(); break;
            case 0xA5: BX_CPU_THIS_PTR SHLD_EvGvCL(); break;
            case 0xA6: BX_CPU_THIS_PTR CMPXCHG_XBTS(); break;
            case 0xA7: BX_CPU_THIS_PTR CMPXCHG_IBTS(); break;
            case 0xA8: BX_CPU_THIS_PTR PUSH_GS(); break;
            case 0xA9: BX_CPU_THIS_PTR POP_GS(); break;
            case 0xAA: bx_panic("decode: : 0F AA:\n"); break;
            case 0xAB: BX_CPU_THIS_PTR BTS_EvGv(); break;
            case 0xAC: BX_CPU_THIS_PTR SHRD_EvGvIb(); break;
            case 0xAD: BX_CPU_THIS_PTR SHRD_EvGvCL(); break;
            case 0xAE: bx_panic("decode: 0F AE:\n"); break;
            case 0xAF: BX_CPU_THIS_PTR IMUL_GvEv(); break;

            case 0xB0: BX_CPU_THIS_PTR CMPXCHG_EbGb(); break;
            case 0xB1: BX_CPU_THIS_PTR CMPXCHG_EvGv(); break;
            case 0xB2: BX_CPU_THIS_PTR LSS_GvMp(); break;
            case 0xB3: BX_CPU_THIS_PTR BTR_EvGv(); break;
            case 0xB4: BX_CPU_THIS_PTR LFS_GvMp(); break;
            case 0xB5: BX_CPU_THIS_PTR LGS_GvMp(); break;
            case 0xB6: BX_CPU_THIS_PTR MOVZX_GvEb(); break;
            case 0xB7: BX_CPU_THIS_PTR MOVZX_GvEw(); break;
            case 0xB8: bx_panic("decode: 0F B8:\n"); break;
            case 0xB9: bx_panic("decode: 0F B9:\n"); break;
            case 0xBA: /* Group 8 Ev,Ib */
              mod_rm_byte = peek_next_byte();
              BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
              switch (opcode) {
                case 0x00:
                case 0x01:
                case 0x02:
                case 0x03: bx_panic("decode: Grp8:\n"); break;
                case 0x04: BX_CPU_THIS_PTR BT_EvIb(); break;
                case 0x05: BX_CPU_THIS_PTR BTS_EvIb(); break;
                case 0x06: BX_CPU_THIS_PTR BTR_EvIb(); break;
                case 0x07: BX_CPU_THIS_PTR BTC_EvIb(); break;
                default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
                }
              break;

            case 0xBB: BX_CPU_THIS_PTR BTC_EvGv(); break;
            case 0xBC: BX_CPU_THIS_PTR BSF_GvEv(); break;
            case 0xBD: BX_CPU_THIS_PTR BSR_GvEv(); break;
            case 0xBE: BX_CPU_THIS_PTR MOVSX_GvEb(); break;
            case 0xBF: BX_CPU_THIS_PTR MOVSX_GvEw(); break;

            case 0xC0: BX_CPU_THIS_PTR XADD_EbGb(); break;
            case 0xC1: BX_CPU_THIS_PTR XADD_EvGv(); break;
            case 0xC2: 
            case 0xC3: 
            case 0xC4: 
            case 0xC5: 
            case 0xC6: 
            case 0xC7: bx_panic("decode: 0F C7:\n"); break;
            case 0xC8: BX_CPU_THIS_PTR BSWAP_EAX(); break;
            case 0xC9: BX_CPU_THIS_PTR BSWAP_ECX(); break;
            case 0xCA: BX_CPU_THIS_PTR BSWAP_EDX(); break;
            case 0xCB: BX_CPU_THIS_PTR BSWAP_EBX(); break;
            case 0xCC: BX_CPU_THIS_PTR BSWAP_ESP(); break;
            case 0xCD: BX_CPU_THIS_PTR BSWAP_EBP(); break;
            case 0xCE: BX_CPU_THIS_PTR BSWAP_ESI(); break;
            case 0xCF: BX_CPU_THIS_PTR BSWAP_EDI(); break;

            case 0xD0:
            case 0xD1:
            case 0xD2:
            case 0xD3:
            case 0xD4:
            case 0xD5:
            case 0xD6:
            case 0xD7:
            case 0xD8:
            case 0xD9:
            case 0xDA:
            case 0xDB:
            case 0xDC:
            case 0xDD:
            case 0xDE:
            case 0xDF:
              bx_panic("decode: 0F %02x:\n", (unsigned) next_byte);
              break;

            case 0xE0:
            case 0xE1:
            case 0xE2:
            case 0xE3:
            case 0xE4:
            case 0xE5:
            case 0xE6:
            case 0xE7:
              bx_panic("decode: 0F %02x:\n", (unsigned) next_byte);
              break;
            case 0xE8:
              bx_panic("decode: 0F %02x: caused undefined opcode exception\n",
                       (unsigned) next_byte);
              exception(6, 0, 0);
              break;
            case 0xE9:
            case 0xEA:
            case 0xEB:
            case 0xEC:
            case 0xED:
            case 0xEE:
            case 0xEF:
              bx_panic("decode: 0F %02x:\n", (unsigned) next_byte);
              exception(6, 0, 0);
              break;

            case 0xF0:
            case 0xF1:
            case 0xF2:
            case 0xF3:
            case 0xF4:
            case 0xF5:
            case 0xF6:
            case 0xF7:
            case 0xF8:
            case 0xF9:
            case 0xFA:
            case 0xFB:
            case 0xFC:
            case 0xFD:
            case 0xFE:
            case 0xFF:
              bx_printf("decode: instruction 0F Fx causes exception 6\n");
              exception(6, 0, 0);
              break;

            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0x10: BX_CPU_THIS_PTR ADC_EbGb(); break;
        case 0x11: BX_CPU_THIS_PTR ADC_EvGv(); break;
        case 0x12: BX_CPU_THIS_PTR ADC_GbEb(); break;
        case 0x13: BX_CPU_THIS_PTR ADC_GvEv(); break;
        case 0x14: BX_CPU_THIS_PTR ADC_ALIb(); break;
        case 0x15: BX_CPU_THIS_PTR ADC_eAXIv(); break;
        case 0x16: BX_CPU_THIS_PTR PUSH_SS(); break;
        case 0x17: BX_CPU_THIS_PTR POP_SS(); break;
        case 0x18: BX_CPU_THIS_PTR SBB_EbGb(); break;
        case 0x19: BX_CPU_THIS_PTR SBB_EvGv(); break;
        case 0x1A: BX_CPU_THIS_PTR SBB_GbEb(); break;
        case 0x1B: BX_CPU_THIS_PTR SBB_GvEv(); break;
        case 0x1C: BX_CPU_THIS_PTR SBB_ALIb(); break;
        case 0x1D: BX_CPU_THIS_PTR SBB_eAXIv(); break;
        case 0x1E: BX_CPU_THIS_PTR PUSH_DS(); break;
        case 0x1F: BX_CPU_THIS_PTR POP_DS(); break;


        case 0x20: BX_CPU_THIS_PTR AND_EbGb(); break;
        case 0x21: BX_CPU_THIS_PTR AND_EvGv(); break;
        case 0x22: BX_CPU_THIS_PTR AND_GbEb(); break;
        case 0x23: BX_CPU_THIS_PTR AND_GvEv(); break;
        case 0x24: BX_CPU_THIS_PTR AND_ALIb(); break;
        case 0x25: BX_CPU_THIS_PTR AND_eAXIv(); break;
        case 0x26: 
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR es;
          continue;
          break;
        case 0x27: BX_CPU_THIS_PTR DAA(); break;
        case 0x28: BX_CPU_THIS_PTR SUB_EbGb(); break;
        case 0x29: BX_CPU_THIS_PTR SUB_EvGv(); break;
        case 0x2A: BX_CPU_THIS_PTR SUB_GbEb(); break;
        case 0x2B: BX_CPU_THIS_PTR SUB_GvEv(); break;
        case 0x2C: BX_CPU_THIS_PTR SUB_ALIb(); break;
        case 0x2D: BX_CPU_THIS_PTR SUB_eAXIv(); break;
        case 0x2E:
#if BX_CPU <= 2
          bx_panic("decode: CS prefix used\n");
#else
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR cs;
#endif
          continue;
          break;
        case 0x2F: BX_CPU_THIS_PTR DAS(); break;

        case 0x30: BX_CPU_THIS_PTR XOR_EbGb(); break;
        case 0x31: BX_CPU_THIS_PTR XOR_EvGv(); break;
        case 0x32: BX_CPU_THIS_PTR XOR_GbEb(); break;
        case 0x33: BX_CPU_THIS_PTR XOR_GvEv(); break;
        case 0x34: BX_CPU_THIS_PTR XOR_ALIb(); break;
        case 0x35: BX_CPU_THIS_PTR XOR_eAXIv(); break;
        case 0x36: 
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR ss;
          continue;
          break;
        case 0x37: BX_CPU_THIS_PTR AAA(); break;
        case 0x38: BX_CPU_THIS_PTR CMP_EbGb(); break;
        case 0x39: BX_CPU_THIS_PTR CMP_EvGv(); break;
        case 0x3A: BX_CPU_THIS_PTR CMP_GbEb(); break;
        case 0x3B: BX_CPU_THIS_PTR CMP_GvEv(); break;
        case 0x3C: BX_CPU_THIS_PTR CMP_ALIb(); break;
        case 0x3D: BX_CPU_THIS_PTR CMP_eAXIv(); break;
        case 0x3E: 
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR ds;
          continue;
          break;
        case 0x3F: BX_CPU_THIS_PTR AAS(); break;

        case 0x40: BX_CPU_THIS_PTR INC_eAX(); break;
        case 0x41: BX_CPU_THIS_PTR INC_eCX(); break;
        case 0x42: BX_CPU_THIS_PTR INC_eDX(); break;
        case 0x43: BX_CPU_THIS_PTR INC_eBX(); break;
        case 0x44: BX_CPU_THIS_PTR INC_eSP(); break;
        case 0x45: BX_CPU_THIS_PTR INC_eBP(); break;
        case 0x46: BX_CPU_THIS_PTR INC_eSI(); break;
        case 0x47: BX_CPU_THIS_PTR INC_eDI(); break;
        case 0x48: BX_CPU_THIS_PTR DEC_eAX(); break;
        case 0x49: BX_CPU_THIS_PTR DEC_eCX(); break;
        case 0x4A: BX_CPU_THIS_PTR DEC_eDX(); break;
        case 0x4B: BX_CPU_THIS_PTR DEC_eBX(); break;
        case 0x4C: BX_CPU_THIS_PTR DEC_eSP(); break;
        case 0x4D: BX_CPU_THIS_PTR DEC_eBP(); break;
        case 0x4E: BX_CPU_THIS_PTR DEC_eSI(); break;
        case 0x4F: BX_CPU_THIS_PTR DEC_eDI(); break;


        case 0x50: BX_CPU_THIS_PTR PUSH_eAX(); break;
        case 0x51: BX_CPU_THIS_PTR PUSH_eCX(); break;
        case 0x52: BX_CPU_THIS_PTR PUSH_eDX(); break;
        case 0x53: BX_CPU_THIS_PTR PUSH_eBX(); break;
        case 0x54: BX_CPU_THIS_PTR PUSH_eSP(); break;
        case 0x55: BX_CPU_THIS_PTR PUSH_eBP(); break;
        case 0x56: BX_CPU_THIS_PTR PUSH_eSI(); break;
        case 0x57: BX_CPU_THIS_PTR PUSH_eDI(); break;
        case 0x58: BX_CPU_THIS_PTR POP_eAX(); break;
        case 0x59: BX_CPU_THIS_PTR POP_eCX(); break;
        case 0x5A: BX_CPU_THIS_PTR POP_eDX(); break;
        case 0x5B: BX_CPU_THIS_PTR POP_eBX(); break;
        case 0x5C: BX_CPU_THIS_PTR POP_eSP(); break;
        case 0x5D: BX_CPU_THIS_PTR POP_eBP(); break;
        case 0x5E: BX_CPU_THIS_PTR POP_eSI(); break;
        case 0x5F: BX_CPU_THIS_PTR POP_eDI(); break;


        case 0x60: BX_CPU_THIS_PTR PUSHAD(); break;
        case 0x61: BX_CPU_THIS_PTR POPAD(); break;
        case 0x62: BX_CPU_THIS_PTR BOUND_GvMa(); break;
        case 0x63: BX_CPU_THIS_PTR ARPL_EwGw(); break;
        case 0x64: 
#if BX_CPU <= 2
          bx_panic("decode: FS segment override used\n");
#else
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR fs;
#endif
          continue;
          break;
        case 0x65: 
#if BX_CPU <= 2
          bx_panic("decode: GS segment override used\n");
#else
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR gs;
#endif
          continue;
          break;
        case 0x66:
#if BX_CPU < 3
          bx_panic("decode: 32bit_opsize prefix not supported on < 386\n");
          exception(6, 0, 0);
#else /* 386+ */
          BX_CPU_THIS_PTR is_32bit_opsize   = !BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
#endif
          continue;
          break;
        case 0x67:
#if BX_CPU < 3
          bx_panic("decode: 32bit_addrsize prefix not supported < 386!\n");
          exception(6, 0, 0);
#else /* 386+ */
          BX_CPU_THIS_PTR is_32bit_addrsize = !BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
#endif
          continue;
          break;
        case 0x68: BX_CPU_THIS_PTR PUSH_Iv(); break;
        case 0x69: BX_CPU_THIS_PTR IMUL_GvEvIv(); break;
        case 0x6A: BX_CPU_THIS_PTR PUSH_Ib(); break;
        case 0x6B: BX_CPU_THIS_PTR IMUL_GvEvIb(); break;
        case 0x6C:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR INSB_YbDX();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR INSB_YbDX();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR INSB_YbDX();
          break;


        case 0x6D:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR INSW_YvDX();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR INSW_YvDX();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR INSW_YvDX();
          break;

        case 0x6E:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR OUTSB_DXXb();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR OUTSB_DXXb();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR OUTSB_DXXb();
          break;


        case 0x6F:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR OUTSW_DXXv();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR OUTSW_DXXv();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR OUTSW_DXXv();
          break;


        case 0x70: BX_CPU_THIS_PTR JO_Jb(); break;
        case 0x71: BX_CPU_THIS_PTR JNO_Jb(); break;
        case 0x72: BX_CPU_THIS_PTR JB_Jb(); break;
        case 0x73: BX_CPU_THIS_PTR JNB_Jb(); break;
        case 0x74: BX_CPU_THIS_PTR JZ_Jb(); break;
        case 0x75: BX_CPU_THIS_PTR JNZ_Jb(); break;
        case 0x76: BX_CPU_THIS_PTR JBE_Jb(); break;
        case 0x77: BX_CPU_THIS_PTR JNBE_Jb(); break;
        case 0x78: BX_CPU_THIS_PTR JS_Jb(); break;
        case 0x79: BX_CPU_THIS_PTR JNS_Jb(); break;
        case 0x7A: BX_CPU_THIS_PTR JP_Jb(); break;
        case 0x7B: BX_CPU_THIS_PTR JNP_Jb(); break;
        case 0x7C: BX_CPU_THIS_PTR JL_Jb(); break;
        case 0x7D: BX_CPU_THIS_PTR JNL_Jb(); break;
        case 0x7E: BX_CPU_THIS_PTR JLE_Jb(); break;
        case 0x7F: BX_CPU_THIS_PTR JNLE_Jb(); break;

        case 0x80: /* Immdediate Group 1 EbIb */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ADD_EbIb(); break;
            case 0x01: BX_CPU_THIS_PTR OR_EbIb();  break;
            case 0x02: BX_CPU_THIS_PTR ADC_EbIb(); break;
            case 0x03: BX_CPU_THIS_PTR SBB_EbIb(); break;
            case 0x04: BX_CPU_THIS_PTR AND_EbIb(); break;
            case 0x05: BX_CPU_THIS_PTR SUB_EbIb(); break;
            case 0x06: BX_CPU_THIS_PTR XOR_EbIb(); break;
            case 0x07: BX_CPU_THIS_PTR CMP_EbIb(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0x81: /* Immdediate Group 1 EvIv */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ADD_EvIv(); break;
            case 0x01: BX_CPU_THIS_PTR OR_EvIv(); break;
            case 0x02: BX_CPU_THIS_PTR ADC_EvIv(); break;
            case 0x03: BX_CPU_THIS_PTR SBB_EvIv(); break;
            case 0x04: BX_CPU_THIS_PTR AND_EvIv(); break;
            case 0x05: BX_CPU_THIS_PTR SUB_EvIv(); break;
            case 0x06: BX_CPU_THIS_PTR XOR_EvIv(); break;
            case 0x07: BX_CPU_THIS_PTR CMP_EvIv(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0x82: bx_panic("decode: 82:\n"); break;

        case 0x83: /* Immdediate Group 1 EvIb */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ADD_EvIb(); break;
            case 0x01: BX_CPU_THIS_PTR OR_EvIb(); break;
            case 0x02: BX_CPU_THIS_PTR ADC_EvIb(); break;
            case 0x03: BX_CPU_THIS_PTR SBB_EvIb(); break;
            case 0x04: BX_CPU_THIS_PTR AND_EvIb(); break;
            case 0x05: BX_CPU_THIS_PTR SUB_EvIb(); break;
            case 0x06: BX_CPU_THIS_PTR XOR_EvIb(); break;
            case 0x07: BX_CPU_THIS_PTR CMP_EvIb(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0x84: BX_CPU_THIS_PTR TEST_EbGb(); break;
        case 0x85: BX_CPU_THIS_PTR TEST_EvGv(); break;
        case 0x86: BX_CPU_THIS_PTR XCHG_EbGb(); break;
        case 0x87: BX_CPU_THIS_PTR XCHG_EvGv(); break;
        case 0x88: BX_CPU_THIS_PTR MOV_EbGb(); break;
        case 0x89: BX_CPU_THIS_PTR MOV_EvGv(); break;
        case 0x8A: BX_CPU_THIS_PTR MOV_GbEb(); break;
        case 0x8B: BX_CPU_THIS_PTR MOV_GvEv(); break;
        case 0x8C: BX_CPU_THIS_PTR MOV_EwSw(); break;
        case 0x8D: BX_CPU_THIS_PTR LEA_GvM(); break;
        case 0x8E: BX_CPU_THIS_PTR MOV_SwEw(); break;
        case 0x8F: BX_CPU_THIS_PTR POP_Ev(); break;


        case 0x90: /* NOP */ break;
        case 0x91: BX_CPU_THIS_PTR XCHG_eCXeAX(); break;
        case 0x92: BX_CPU_THIS_PTR XCHG_eDXeAX(); break;
        case 0x93: BX_CPU_THIS_PTR XCHG_eBXeAX(); break;
        case 0x94: BX_CPU_THIS_PTR XCHG_eSPeAX(); break;
        case 0x95: BX_CPU_THIS_PTR XCHG_eBPeAX(); break;
        case 0x96: BX_CPU_THIS_PTR XCHG_eSIeAX(); break;
        case 0x97: BX_CPU_THIS_PTR XCHG_eDIeAX(); break;
        case 0x98: BX_CPU_THIS_PTR CBW(); break;
        case 0x99: BX_CPU_THIS_PTR CWD(); break;
        case 0x9A: BX_CPU_THIS_PTR CALL_Ap(); break;
        case 0x9B: BX_CPU_THIS_PTR WAIT(); break;
        case 0x9C: BX_CPU_THIS_PTR PUSHF_Fv(); break;
        case 0x9D: BX_CPU_THIS_PTR POPF_Fv(); break;
        case 0x9E: BX_CPU_THIS_PTR SAHF(); break;
        case 0x9F: BX_CPU_THIS_PTR LAHF(); break;


        case 0xA0: BX_CPU_THIS_PTR MOV_ALOb(); break;
        case 0xA1: BX_CPU_THIS_PTR MOV_eAXOv(); break;
        case 0xA2: BX_CPU_THIS_PTR MOV_ObAL(); break;
        case 0xA3: BX_CPU_THIS_PTR MOV_OveAX(); break;
        case 0xA4:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR MOVSB_XbYb();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR MOVSB_XbYb();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR MOVSB_XbYb();
          break;

        case 0xA5:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR MOVSW_XvYv();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR MOVSW_XvYv();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR MOVSW_XvYv();
          break;

        case 0xA6:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR CMPSB_XbYb();
                ECX = ECX - 1;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR CMPSB_XbYb();
                CX = CX - 1;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR CMPSB_XbYb();
          break;

        case 0xA7:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR CMPSW_XvYv();
                ECX = ECX - 1;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR CMPSW_XvYv();
                CX = CX - 1;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR CMPSW_XvYv();
          break;

        case 0xA8: BX_CPU_THIS_PTR TEST_ALIb(); break;
        case 0xA9: BX_CPU_THIS_PTR TEST_eAXIv(); break;
        case 0xAA:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR STOSB_YbAL();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR STOSB_YbAL();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR STOSB_YbAL();
          break;

        case 0xAB:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR STOSW_YveAX();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR STOSW_YveAX();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR STOSW_YveAX();
          break;

        case 0xAC:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR LODSB_ALXb();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR LODSB_ALXb();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR LODSB_ALXb();
          break;

        case 0xAD:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR LODSW_eAXXv();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR LODSW_eAXXv();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR LODSW_eAXXv();
          break;

        case 0xAE:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR SCASB_ALXb();
                ECX = ECX - 1;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR SCASB_ALXb();
                CX = CX - 1;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR SCASB_ALXb();
          break;

        case 0xAF:
          if (BX_CPU_THIS_PTR rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                BX_CPU_THIS_PTR SCASW_eAXXv();
                ECX = ECX - 1;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                BX_CPU_THIS_PTR SCASW_eAXXv();
                CX = CX - 1;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (BX_CPU_THIS_PTR rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          BX_CPU_THIS_PTR SCASW_eAXXv();
          break;


        case 0xB0: BX_CPU_THIS_PTR MOV_ALIb(); break;
        case 0xB1: BX_CPU_THIS_PTR MOV_CLIb(); break;
        case 0xB2: BX_CPU_THIS_PTR MOV_DLIb(); break;
        case 0xB3: BX_CPU_THIS_PTR MOV_BLIb(); break;
        case 0xB4: BX_CPU_THIS_PTR MOV_AHIb(); break;
        case 0xB5: BX_CPU_THIS_PTR MOV_CHIb(); break;
        case 0xB6: BX_CPU_THIS_PTR MOV_DHIb(); break;
        case 0xB7: BX_CPU_THIS_PTR MOV_BHIb(); break;
        case 0xB8: BX_CPU_THIS_PTR MOV_eAXIv(); break;
        case 0xB9: BX_CPU_THIS_PTR MOV_eCXIv(); break;
        case 0xBA: BX_CPU_THIS_PTR MOV_eDXIv(); break;
        case 0xBB: BX_CPU_THIS_PTR MOV_eBXIv(); break;
        case 0xBC: BX_CPU_THIS_PTR MOV_eSPIv(); break;
        case 0xBD: BX_CPU_THIS_PTR MOV_eBPIv(); break;
        case 0xBE: BX_CPU_THIS_PTR MOV_eSIIv(); break;
        case 0xBF: BX_CPU_THIS_PTR MOV_eDIIv(); break;

        case 0xC0: /* Group 2 Eb,Ib */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ROL_EbIb(); break;
            case 0x01: BX_CPU_THIS_PTR ROR_EbIb(); break;
            case 0x02: BX_CPU_THIS_PTR RCL_EbIb(); break;
            case 0x03: BX_CPU_THIS_PTR RCR_EbIb(); break;
            case 0x04: BX_CPU_THIS_PTR SHL_EbIb(); break;
            case 0x05: BX_CPU_THIS_PTR SHR_EbIb(); break;
            case 0x06: bx_panic("decode: Grp2: (SHL_EbIb duplicate)\n"); break;
            case 0x07: BX_CPU_THIS_PTR SAR_EbIb(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xC1: /* Group 2 Ev,Ib */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ROL_EvIb(); break;
            case 0x01: BX_CPU_THIS_PTR ROR_EvIb(); break;
            case 0x02: BX_CPU_THIS_PTR RCL_EvIb(); break;
            case 0x03: BX_CPU_THIS_PTR RCR_EvIb(); break;
            case 0x04: BX_CPU_THIS_PTR SHL_EvIb(); break;
            case 0x05: BX_CPU_THIS_PTR SHR_EvIb(); break;
            case 0x06: bx_panic("decode: Grp2: (SHL_EvIb duplicate)\n"); break;
            case 0x07: BX_CPU_THIS_PTR SAR_EvIb(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xC2: BX_CPU_THIS_PTR RETnear_Iw(); break;
        case 0xC3: BX_CPU_THIS_PTR RETnear(); break;
        case 0xC4: BX_CPU_THIS_PTR LES_GvMp(); break;
        case 0xC5: BX_CPU_THIS_PTR LDS_GvMp(); break;
        case 0xC6: BX_CPU_THIS_PTR MOV_EbIb(); break;
        case 0xC7: BX_CPU_THIS_PTR MOV_EvIv(); break;
        case 0xC8: BX_CPU_THIS_PTR ENTER_IwIb(); break;
        case 0xC9: BX_CPU_THIS_PTR LEAVE(); break;
        case 0xCA: BX_CPU_THIS_PTR RETfar_Iw(); break;
        case 0xCB: BX_CPU_THIS_PTR RETfar(); break;
        case 0xCC: BX_CPU_THIS_PTR INT3(); break;
        case 0xCD: BX_CPU_THIS_PTR INT_Ib(); break;
        case 0xCE: BX_CPU_THIS_PTR INTO(); break;
        case 0xCF: BX_CPU_THIS_PTR IRET(); break;


        case 0xD0: /* Group 2 Eb,1 */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ROL_Eb1(); break;
            case 0x01: BX_CPU_THIS_PTR ROR_Eb1(); break;
            case 0x02: BX_CPU_THIS_PTR RCL_Eb1(); break;
            case 0x03: BX_CPU_THIS_PTR RCR_Eb1(); break;
            case 0x04: BX_CPU_THIS_PTR SHL_Eb1(); break;
            case 0x05: BX_CPU_THIS_PTR SHR_Eb1(); break;
            case 0x06: bx_panic("decode: Grp2:(SHL_Eb1 duplicate)\n"); break;
            case 0x07: BX_CPU_THIS_PTR SAR_Eb1(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xD1: /* group2 Ev,1 */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ROL_Ev1(); break;
            case 0x01: BX_CPU_THIS_PTR ROR_Ev1(); break;
            case 0x02: BX_CPU_THIS_PTR RCL_Ev1(); break;
            case 0x03: BX_CPU_THIS_PTR RCR_Ev1(); break;
            case 0x04: BX_CPU_THIS_PTR SHL_Ev1(); break;
            case 0x05: BX_CPU_THIS_PTR SHR_Ev1(); break;
            case 0x06: bx_panic("decode: Grp2:(SHL_Ev1 duplicate)\n"); break;
            case 0x07: BX_CPU_THIS_PTR SAR_Ev1(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xD2: /* group2 Eb,CL */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ROL_EbCL(); break;
            case 0x01: BX_CPU_THIS_PTR ROR_EbCL(); break;
            case 0x02: BX_CPU_THIS_PTR RCL_EbCL(); break;
            case 0x03: BX_CPU_THIS_PTR RCR_EbCL(); break;
            case 0x04: BX_CPU_THIS_PTR SHL_EbCL(); break;
            case 0x05: BX_CPU_THIS_PTR SHR_EbCL(); break;
            case 0x06: bx_panic("decode: Grp2:\n"); break;
            case 0x07: BX_CPU_THIS_PTR SAR_EbCL(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xD3: /* group2 Ev,CL */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR ROL_EvCL(); break;
            case 0x01: BX_CPU_THIS_PTR ROR_EvCL(); break;
            case 0x02: BX_CPU_THIS_PTR RCL_EvCL(); break;
            case 0x03: BX_CPU_THIS_PTR RCR_EvCL(); break;
            case 0x04: BX_CPU_THIS_PTR SHL_EvCL(); break;
            case 0x05: BX_CPU_THIS_PTR SHR_EvCL(); break;
            case 0x06: bx_panic("decode: Grp2:\n"); break;
            case 0x07: BX_CPU_THIS_PTR SAR_EvCL(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xD4: BX_CPU_THIS_PTR AAM(); break;
        case 0xD5: BX_CPU_THIS_PTR AAD(); break;
        case 0xD6: bx_panic("decode: D6:(SETALC) Intel reserved opcode\n");
                   break;
        case 0xD7: BX_CPU_THIS_PTR XLAT(); break;

        case 0xD8: /* ESC0 */
          bx_printf("ESC0:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          bx_panic("ESC0:\n");
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          if (mod != 3) {
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FADD_Es(); break;
              case 0x01: BX_CPU_THIS_PTR FMUL_Es(); break;
              case 0x02: BX_CPU_THIS_PTR FCOM_Es(); break;
              case 0x03: BX_CPU_THIS_PTR FCOMP_Es(); break;
              case 0x04: BX_CPU_THIS_PTR FSUB_Es(); break;
              case 0x05: BX_CPU_THIS_PTR FSUBR_Es(); break;
              case 0x06: BX_CPU_THIS_PTR FDIV_Es(); break;
              case 0x07: BX_CPU_THIS_PTR FDIVR_Es(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
/* used peek_next_byte().  do I need to read forward another byte? */
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FADD_ST_STi(); break;
              case 0x01: BX_CPU_THIS_PTR FMUL_ST_STi(); break;
              case 0x02: BX_CPU_THIS_PTR FCOM_ST_STi(); break;
              case 0x03: BX_CPU_THIS_PTR FCOMP_ST_STi(); break;
              case 0x04: BX_CPU_THIS_PTR FSUB_ST_STi(); break;
              case 0x05: BX_CPU_THIS_PTR FSUBR_ST_STi(); break;
              case 0x06: BX_CPU_THIS_PTR FDIV_ST_STi(); break;
              case 0x07: BX_CPU_THIS_PTR FDIVR_ST_STi(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          break;

        case 0xD9: /* ESC1 */
          bx_printf("ESC1:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
bx_printf("decode: ESC1 sequence found, mod = %d, opcode = %d\n",
(int) mod, (int) opcode);
          if (mod != 3) {
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FLD_Es(); break;
              case 0x01: bx_panic("decode: ESC1:\n"); break;
              case 0x02: BX_CPU_THIS_PTR FST_Es(); break;
              case 0x03: BX_CPU_THIS_PTR FSTP_Es(); break;
              case 0x04: BX_CPU_THIS_PTR FLDENV_Ea(); break;
              case 0x05: BX_CPU_THIS_PTR FLDCW_Ew(); break;
              case 0x06: BX_CPU_THIS_PTR FSTENV_Ea(); break;
              case 0x07: BX_CPU_THIS_PTR FSTCW_Ew(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
/* used peek_next_byte().  do I need to read forward another byte? */
            switch (opcode) {
              case 0x00:
                BX_CPU_THIS_PTR FLD_STi(); break;
              case 0x01:
                BX_CPU_THIS_PTR FXCH_STi(); break;
              case 0x02:
                if (rm == 0) {
                  BX_CPU_THIS_PTR FNOP(); break;
                  }
                else {
                  bx_panic("decode: ESC9:\n"); break;
                  }
              case 0x03:
                BX_CPU_THIS_PTR FSTP_STi(); break;

              case 0x04:
                switch (rm) {
                  case 0x00: BX_CPU_THIS_PTR FCHS(); break;
                  case 0x01: BX_CPU_THIS_PTR FABS(); break;
                  case 0x02:
                  case 0x03: bx_panic("decode: ESC9:\n"); break;
                  case 0x04: BX_CPU_THIS_PTR FTST(); break;
                  case 0x05: BX_CPU_THIS_PTR FXAM(); break;
                  case 0x06:
                  case 0x07: bx_panic("decode: ESC9:\n"); break;
                  }
                break;
              case 0x05:
                switch (rm) {
                  case 0x00: BX_CPU_THIS_PTR FLD1(); break;
                  case 0x01: BX_CPU_THIS_PTR FLDL2T(); break;
                  case 0x02: BX_CPU_THIS_PTR FLDL2E(); break;
                  case 0x03: BX_CPU_THIS_PTR FLDPI(); break;
                  case 0x04: BX_CPU_THIS_PTR FLDLG2(); break;
                  case 0x05: BX_CPU_THIS_PTR FLDLN2(); break;
                  case 0x06: BX_CPU_THIS_PTR FLDZ(); break;
                  case 0x07: bx_panic("decode: ESC9:\n"); break;
                  }
                break;
              case 0x06:
                switch (rm) {
                  case 0x00: BX_CPU_THIS_PTR F2XM1(); break;
                  case 0x01: BX_CPU_THIS_PTR FYL2X(); break;
                  case 0x02: BX_CPU_THIS_PTR FPTAN(); break;
                  case 0x03: BX_CPU_THIS_PTR FPATAN(); break;
                  case 0x04: BX_CPU_THIS_PTR FXTRACT(); break;
                  case 0x05: BX_CPU_THIS_PTR FPREM1(); break;
                  case 0x06: BX_CPU_THIS_PTR FDECSTP(); break;
                  case 0x07: BX_CPU_THIS_PTR FINCSTP(); break;
                  }
                break;
              case 0x07:
                switch (rm) {
                  case 0x00: BX_CPU_THIS_PTR FPREM(); break;
                  case 0x01: BX_CPU_THIS_PTR FYL2XP1(); break;
                  case 0x02: BX_CPU_THIS_PTR FSQRT(); break;
                  case 0x03: BX_CPU_THIS_PTR FSINCOS(); break;
                  case 0x04: BX_CPU_THIS_PTR FRNDINT(); break;
                  case 0x05: BX_CPU_THIS_PTR FSCALE(); break;
                  case 0x06: BX_CPU_THIS_PTR FSIN(); break;
                  case 0x07: BX_CPU_THIS_PTR FCOS(); break;
                  }
                break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          break;

        case 0xDA: /* ESC2 */
          bx_printf("ESC2:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          bx_panic("ESC2:\n");
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          if (mod != 3) {
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FIADD_Ed(); break;
              case 0x01: BX_CPU_THIS_PTR FIMUL_Ed(); break;
              case 0x02: BX_CPU_THIS_PTR FICOM_Ed(); break;
              case 0x03: BX_CPU_THIS_PTR FICOMP_Ed(); break;
              case 0x04: BX_CPU_THIS_PTR FISUB_Ed(); break;
              case 0x05: BX_CPU_THIS_PTR FISUBR_Ed(); break;
              case 0x06: BX_CPU_THIS_PTR FIDIV_Ed(); break;
              case 0x07: BX_CPU_THIS_PTR FIDIVR_Ed(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
/* used peek_next_byte().  do I need to read forward another byte? */
            switch (opcode) {
              case 0x05:
                if (rm == 1) {
                  BX_CPU_THIS_PTR FUCOMPP(); break;
                  }
                else {
                  bx_panic("decode: ESC2:\n"); break;
                  }
              default: bx_panic("decode: ESC2:\n"); break;
              }
            break;
            }
          break;

        case 0xDB: /* ESC3 */
          bx_printf("ESC3:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
#if 0
bx_panic("decode: ESC3 sequence found, mod = %d, opcode = %d\n",
(int) mod, (int) opcode);
#endif
          if (mod != 3) {
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FILD_Ed(); break;
              case 0x01: bx_panic("decode: ESC3:\n"); break;
              case 0x02: BX_CPU_THIS_PTR FIST_Ed(); break;
              case 0x03: BX_CPU_THIS_PTR FISTP_Ed(); break;
              case 0x04: bx_panic("decode: ESC3:\n"); break;
              case 0x05: BX_CPU_THIS_PTR FLD_Et(); break;
              case 0x06: bx_panic("decode: ESC3:\n"); break;
              case 0x07: BX_CPU_THIS_PTR FSTP_Et(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
            (void) fetch_next_byte();
            switch (mod_rm_byte) {
              case 0xe0: BX_CPU_THIS_PTR FENI(); break;
              case 0xe1: BX_CPU_THIS_PTR FDISI(); break;
              case 0xe2: BX_CPU_THIS_PTR FCLEX(); break;
              case 0xe3: BX_CPU_THIS_PTR FINIT(); break;
              case 0xe4: BX_CPU_THIS_PTR FSETPM(); break;
              default:
                bx_panic("decode: ESC3: reserved opcode\n"); break;
              }
            break;
            }
          break;

        case 0xDC: /* ESC4 */
          bx_printf("ESC4:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          bx_panic("ESC4:\n");
          mod_rm_byte = peek_next_byte();
/* used peek_next_byte().  do I need to read forward another byte? */
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          if (mod != 3) {
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FADD_El(); break;
              case 0x01: BX_CPU_THIS_PTR FMUL_El(); break;
              case 0x02: BX_CPU_THIS_PTR FCOM_El(); break;
              case 0x03: BX_CPU_THIS_PTR FCOMP_El(); break;
              case 0x04: BX_CPU_THIS_PTR FSUB_El(); break;
              case 0x05: BX_CPU_THIS_PTR FSUBR_El(); break;
              case 0x06: BX_CPU_THIS_PTR FDIV_El(); break;
              case 0x07: BX_CPU_THIS_PTR FDIVR_El(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
/* used peek_next_byte().  do I need to read forward another byte? */
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FADD_STi_ST(); break;
              case 0x01: BX_CPU_THIS_PTR FMUL_STi_ST(); break;
              case 0x02: BX_CPU_THIS_PTR FCOM_STi_ST(); break;
              case 0x03: BX_CPU_THIS_PTR FCOMP_STi_ST(); break;
              case 0x04: BX_CPU_THIS_PTR FSUBR_STi_ST(); break;
              case 0x05: BX_CPU_THIS_PTR FSUB_STi_ST(); break;
              case 0x06: BX_CPU_THIS_PTR FDIVR_STi_ST(); break;
              case 0x07: BX_CPU_THIS_PTR FDIV_STi_ST(); break;
              default: bx_panic("decode: ESC4:\n"); break;
              }
            break;
            }
          break;


        case 0xDD: /* ESC5 */
          bx_printf("ESC5:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          if (mod != 3) {
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FLD_El(); break;
              case 0x01: bx_panic("decode: ESC5:\n"); break;
              case 0x02: BX_CPU_THIS_PTR FST_El(); break;
              case 0x03: BX_CPU_THIS_PTR FSTP_El(); break;
              case 0x04: BX_CPU_THIS_PTR FRSTOR_Ea(); break;
              case 0x05: bx_panic("decode: ESC5:\n"); break;
              case 0x06: BX_CPU_THIS_PTR FSAVE_Ea(); break;
              case 0x07: BX_CPU_THIS_PTR FSTSW_Ew(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
bx_panic("ESC5: recode\n");
/* used peek_next_byte().  do I need to read forward another byte? */
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FFREE_STi(); break;
              case 0x01: BX_CPU_THIS_PTR FXCH_STi(); break;
              case 0x02: BX_CPU_THIS_PTR FST_STi(); break;
              case 0x03: BX_CPU_THIS_PTR FSTP_STi(); break;
              case 0x04: BX_CPU_THIS_PTR FUCOM_STi_ST(); break;
              case 0x05: BX_CPU_THIS_PTR FUCOMP_STi(); break;
              case 0x06: bx_panic("decode: ESC5:\n"); break;
              case 0x07: bx_panic("decode: ESC5:\n"); break;
              default: bx_panic("decode: ESC5:\n"); break;
              }
            break;
            }
          break;

        case 0xDE: /* ESC6 */
          bx_printf("ESC6:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          bx_panic("ESC6:\n");
          mod_rm_byte = peek_next_byte();
/* used peek_next_byte().  do I need to read forward another byte? */
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          if (mod != 3) {
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FIADD_Ew(); break;
              case 0x01: BX_CPU_THIS_PTR FIMUL_Ew(); break;
              case 0x02: BX_CPU_THIS_PTR FICOM_Ew(); break;
              case 0x03: BX_CPU_THIS_PTR FICOMP_Ew(); break;
              case 0x04: BX_CPU_THIS_PTR FISUB_Ew(); break;
              case 0x05: BX_CPU_THIS_PTR FISUBR_Ew(); break;
              case 0x06: BX_CPU_THIS_PTR FIDIV_Ew(); break;
              case 0x07: BX_CPU_THIS_PTR FIDIVR_Ew(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
/* used peek_next_byte().  do I need to read forward another byte? */
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FADDP_STi_ST(); break;
              case 0x01: BX_CPU_THIS_PTR FMULP_STi_ST(); break;
              case 0x02: BX_CPU_THIS_PTR FCOMP_STi(); break;
              case 0x03:
                switch (rm) {
                  case 0x01: BX_CPU_THIS_PTR FCOMPP(); break;
                  default: bx_panic("decode: ESC6:\n"); break;
                  }
                break;
              case 0x04: BX_CPU_THIS_PTR FSUBRP_STi_ST(); break;
              case 0x05: BX_CPU_THIS_PTR FSUBP_STi_ST(); break;
              case 0x06: BX_CPU_THIS_PTR FDIVRP_STi_ST(); break;
              case 0x07: BX_CPU_THIS_PTR FDIVP_STi_ST(); break;
              default: bx_panic("decode: ESC6:\n"); break;
              }
            break;
            }
          break;

        case 0xDF: /* ESC7 */
          bx_printf("ESC7:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          if (mod != 3) {
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FILD_Ew(); break;
              case 0x01: bx_panic("decode: ESC7: mod!=3 case 0x01:\n"); break;
              case 0x02: BX_CPU_THIS_PTR FIST_Ew(); break;
              case 0x03: BX_CPU_THIS_PTR FISTP_Ew(); break;
              case 0x04: BX_CPU_THIS_PTR FBLD_Eb(); break;
              case 0x05: BX_CPU_THIS_PTR FILD_Eq(); break;
              case 0x06: BX_CPU_THIS_PTR FBSTP_Eb(); break;
              case 0x07: BX_CPU_THIS_PTR FISTP_Eq(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
/* used peek_next_byte().  do I need to read forward another byte? */
            switch (opcode) {
              case 0x00: BX_CPU_THIS_PTR FFREE_STi(); break;
              case 0x01: BX_CPU_THIS_PTR FXCH_STi(); break;
              case 0x02: BX_CPU_THIS_PTR FST_STi(); break;
              case 0x03: BX_CPU_THIS_PTR FSTP_STi(); break;
              case 0x04:
                switch (rm) {
                  case 0x00: BX_CPU_THIS_PTR FSTSW_AX(); break;
                  default: bx_panic("decode: ESC7: mod==3 case 0x04: default:\n"); break;
                  }
                break;
              default: bx_panic("decode: ESC7: mod==3: default\n"); break;
              }
            break;
            }
          break;


        case 0xE0: BX_CPU_THIS_PTR LOOPNE_Jb(); break;
        case 0xE1: BX_CPU_THIS_PTR LOOPE_Jb(); break;
        case 0xE2: BX_CPU_THIS_PTR LOOP_Jb(); break;
        case 0xE3: BX_CPU_THIS_PTR JCXZ_Jb(); break;
        case 0xE4: BX_CPU_THIS_PTR IN_ALIb(); break;
        case 0xE5: BX_CPU_THIS_PTR IN_eAXIb(); break;
        case 0xE6: BX_CPU_THIS_PTR OUT_IbAL(); break;
        case 0xE7: BX_CPU_THIS_PTR OUT_IbeAX(); break;
        case 0xE8: BX_CPU_THIS_PTR CALL_Av(); break;
        case 0xE9: BX_CPU_THIS_PTR JMP_Jv(); break;
        case 0xEA: BX_CPU_THIS_PTR JMP_Ap(); break;
        case 0xEB: BX_CPU_THIS_PTR JMP_Jb(); break;
        case 0xEC: BX_CPU_THIS_PTR IN_ALDX(); break;
        case 0xED: BX_CPU_THIS_PTR IN_eAXDX(); break;
        case 0xEE: BX_CPU_THIS_PTR OUT_DXAL(); break;
        case 0xEF: BX_CPU_THIS_PTR OUT_DXeAX(); break;

        case 0xF0: /* LOCK */
          bx_printf("decode: LOCK prefix encountered\n");
          if (protected_mode() && CPL>IOPL) {
            bx_panic("decode: LOCK prefix found: CPL>IOPL\n");
            /* #GP(0) */
            break;
            }
          continue;
          break;
        case 0xF1: bx_panic("decode: F1: Intel reserved opcode\n");
                   break;
        case 0xF2: /* REPNE/REPNZ */
          BX_CPU_THIS_PTR rep_prefix = BX_REPNE_PREFIX;
          continue;
          break;
        case 0xF3: /* REP/REPE/REPZ */
          BX_CPU_THIS_PTR rep_prefix = BX_REPE_PREFIX;
          continue;
          break;
        case 0xF4:
          BX_CPU_THIS_PTR HLT();
          break;
        case 0xF5: BX_CPU_THIS_PTR CMC(); break;
        case 0xF6: /* Group 3 Eb */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR TEST_EbIb(); break;
            case 0x01: bx_panic("decode: F6:TEST_EbIb duplicate\n"); break;
            case 0x02: BX_CPU_THIS_PTR NOT_Eb(); break;
            case 0x03: BX_CPU_THIS_PTR NEG_Eb(); break;
            case 0x04: BX_CPU_THIS_PTR MUL_ALEb(); break;
            case 0x05: BX_CPU_THIS_PTR IMUL_ALEb(); break;
            case 0x06: BX_CPU_THIS_PTR DIV_ALEb(); break;
            case 0x07: BX_CPU_THIS_PTR IDIV_ALEb(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xF7: /* GROUP3 Ev */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR TEST_EvIv(); break;
            case 0x01: bx_panic("decode: F7 01:TEST_EvIv duplicate\n"); break;
            case 0x02: BX_CPU_THIS_PTR NOT_Ev(); break;
            case 0x03: BX_CPU_THIS_PTR NEG_Ev(); break;
            case 0x04: BX_CPU_THIS_PTR MUL_eAXEv(); break;
            case 0x05: BX_CPU_THIS_PTR IMUL_eAXEv(); break;
            case 0x06: BX_CPU_THIS_PTR DIV_eAXEv(); break;
            case 0x07: BX_CPU_THIS_PTR IDIV_eAXEv(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;
        case 0xF8: BX_CPU_THIS_PTR CLC(); break;
        case 0xF9: BX_CPU_THIS_PTR STC(); break;
        case 0xFA: BX_CPU_THIS_PTR CLI(); break;
        case 0xFB: BX_CPU_THIS_PTR STI(); break;
        case 0xFC: BX_CPU_THIS_PTR CLD(); break;
        case 0xFD: BX_CPU_THIS_PTR STD(); break;
        case 0xFE: /* GROUP4 */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR INC_Eb(); break;
            case 0x01: BX_CPU_THIS_PTR DEC_Eb(); break;
            default: bx_panic("decode: FE default:\n"); break;
            }
          break;

        case 0xFF: /* GROUP 5 */
          mod_rm_byte = peek_next_byte();
          BX_DECODE_MODRM(mod_rm_byte, mod, opcode, rm);
          switch (opcode) {
            case 0x00: BX_CPU_THIS_PTR INC_Ev(); break;
            case 0x01: BX_CPU_THIS_PTR DEC_Ev(); break;
            case 0x02: BX_CPU_THIS_PTR CALL_Ev(); break;
            case 0x03: BX_CPU_THIS_PTR CALL_Ep(); break;
            case 0x04: BX_CPU_THIS_PTR JMP_Ev(); break;
            case 0x05: BX_CPU_THIS_PTR JMP_Ep(); break;
            case 0x06: BX_CPU_THIS_PTR PUSH_Ev(); break;
            default: bx_panic("decode: FF default:\n"); break;
            }
          break;

        default: /* only invalid instructions left */
          bx_panic("decode: error in bx_cpu_loop()!\n");
          break;
        } /* switch (next_byte) */
      break;
      } // while (1)...

    // advance instruction tick clock...
    bx_pc_system.TICK();
    } /* while (1) */

  BX_CPU_LOOP_EXIT();
}
