// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"





#if BX_SUPPORT_TASKING

#if BX_CPU >= 2
  /* corresponds to SWITCH_TASKS algorithm in Intel documentation */
  void
bx_cpu_c::task_switch16(bx_selector_t *selector,
                 bx_descriptor_t *descriptor, unsigned source,
                 Bit32u dword1, Bit32u dword2)
{
  Bit32u base32;
  Bit16u temp16;
  Bit16u raw_cs_selector, raw_ss_selector, raw_ds_selector, raw_es_selector;
  Bit16u raw_ldt_selector;
  bx_selector_t cs_selector, ss_selector, ds_selector, es_selector,
                ldt_selector;
  bx_descriptor_t cs_descriptor, ss_descriptor, ds_descriptor, es_descriptor,
                  ldt_descriptor;

  if (v8086_mode()) bx_panic("ctrl_xfer_pro: v8086 mode unsupported\n");

  /* ======================
   * 286 Task State Segment
   * ======================
   * offset dynamic item
   * 42     0       task LDT selector
   * 40     1       DS selector
   * 38     1       SS selector
   * 36     1       CS selector
   * 34     1       ES selector
   * 32     1       DI
   * 30     1       SI
   * 28     1       BP
   * 26     1       SP
   * 24     1       BX
   * 22     1       DX
   * 20     1       CX
   * 18     1       AX
   * 16     1       flag word
   * 14     1       IP (entry point)
   * 12     0       SS for CPL 2
   * 10     0       SP for CPL 2
   * 08     0       SS for CPL 1
   * 06     0       SP for CPL 1
   * 04     0       SS for CPL 0
   * 02     0       SP for CPL 0
   * 00             back link selector to TSS
   */

bx_panic("task_switch286: bombing\n");
/* use access_linear or access_physical below? */

/* ??? look at previous errno rather than HANDLE_EXCEPTION */

  /* locked set AR byte of new TSS descriptor to Busy TSS (bit1 = 1) */
  /* ??? what to do upon paging errors if this is changed to
   * access_virtual() */
  dword2 |= 0x00000200;
  access_linear(BX_CPU_THIS_PTR gdtr.base + selector->index*8 + 4, 4, 0,
    BX_WRITE, &dword2);

  /* current TSS cache must be valid with limit >= ???41,
   * else #TS(error code will be new TSS, but back link points at old TSS)*/
   /* ??? back link points to old TSS */
  if (BX_CPU_THIS_PTR tr.cache.valid==0 ||
      BX_CPU_THIS_PTR tr.cache.type!=1 ||
      BX_CPU_THIS_PTR tr.cache.u.tss286.limit < 43) {
    bx_panic("task_switch16(): TR not valid\n");
    exception(BX_TS_EXCEPTION, selector->value & 0xfffc, 0);
    return;
    }

  /* save machine state (dynamic part) in current TSS */
  base32 = BX_CPU_THIS_PTR tr.cache.u.tss286.base; /* current TSS.base */
  temp16 = IP; access_linear(base32 + 14, 2, 0, BX_WRITE, &temp16);
  temp16 = BX_CPU_THIS_PTR read_flags();
               access_linear(base32 + 16, 2, 0, BX_WRITE, &temp16);
  temp16 = AX; access_linear(base32 + 18, 2, 0, BX_WRITE, &temp16);
  temp16 = CX; access_linear(base32 + 20, 2, 0, BX_WRITE, &temp16);
  temp16 = DX; access_linear(base32 + 22, 2, 0, BX_WRITE, &temp16);
  temp16 = BX; access_linear(base32 + 24, 2, 0, BX_WRITE, &temp16);
  temp16 = SP; access_linear(base32 + 26, 2, 0, BX_WRITE, &temp16);
  temp16 = BP; access_linear(base32 + 28, 2, 0, BX_WRITE, &temp16);
  temp16 = SI; access_linear(base32 + 30, 2, 0, BX_WRITE, &temp16);
  temp16 = DI; access_linear(base32 + 32, 2, 0, BX_WRITE, &temp16);
  temp16 = BX_CPU_THIS_PTR es.selector.value;
               access_linear(base32 + 34, 2, 0, BX_WRITE, &temp16);
  temp16 = cs.selector.value;
               access_linear(base32 + 36, 2, 0, BX_WRITE, &temp16);
  temp16 = ss.selector.value;
               access_linear(base32 + 38, 2, 0, BX_WRITE, &temp16);
  temp16 = BX_CPU_THIS_PTR ds.selector.value;
               access_linear(base32 + 40, 2, 0, BX_WRITE, &temp16);

  /* if nesting tasks, set the new TSS link to the current TSS selector */
  if (source == BX_TASK_FROM_CALL_OR_INT) {
    access_linear(descriptor->u.tss286.base + 0, 2, 0, BX_WRITE,
      &BX_CPU_THIS_PTR tr.selector.value);
    }

  /* any exception will be in new context, else set the AR byte of
   * current TSS descriptor to available TSS (Bit1=0) */
  /* ??? */

  /* set the current TR to selector, base, and limit of new TSS */
  BX_CPU_THIS_PTR tr.selector = *selector;
  BX_CPU_THIS_PTR tr.cache    = *descriptor;

  /* new TSS limit >= 43, else #TS(new TSS) */
  if (BX_CPU_THIS_PTR tr.cache.u.tss286.limit < 43) {
    bx_panic("task_switch16: new TSS limit < 43\n");
    exception(BX_TS_EXCEPTION, selector->value & 0xfffc, 0);
    return;
    }

  /* set all machine registers to value from new TSS without loading
   * descriptors for DS, ES, CS, SS, LDT */
  base32 = BX_CPU_THIS_PTR tr.cache.u.tss286.base;
  access_linear(base32 + 14, 2, 0, BX_READ, &temp16); EIP = temp16;
  access_linear(base32 + 16, 2, 0, BX_READ, &temp16);
    write_flags(temp16, 1, 1);
  access_linear(base32 + 18, 2, 0, BX_READ, &temp16); AX = temp16;
  access_linear(base32 + 20, 2, 0, BX_READ, &temp16); CX = temp16;
  access_linear(base32 + 22, 2, 0, BX_READ, &temp16); DX = temp16;
  access_linear(base32 + 24, 2, 0, BX_READ, &temp16); BX = temp16;
  access_linear(base32 + 26, 2, 0, BX_READ, &temp16); ESP = temp16;
  access_linear(base32 + 28, 2, 0, BX_READ, &temp16); BP = temp16;
  access_linear(base32 + 30, 2, 0, BX_READ, &temp16); SI = temp16;
  access_linear(base32 + 32, 2, 0, BX_READ, &temp16); DI = temp16;

  access_linear(base32 + 34, 2, 0, BX_READ, &raw_es_selector);
  access_linear(base32 + 36, 2, 0, BX_READ, &raw_cs_selector);
  access_linear(base32 + 38, 2, 0, BX_READ, &raw_ss_selector);
  access_linear(base32 + 40, 2, 0, BX_READ, &raw_ds_selector);
  access_linear(base32 + 42, 2, 0, BX_READ, &raw_ldt_selector);


  /* clear valid flags for LDT, SS, CS, DS, ES (not valid yet) */
  BX_CPU_THIS_PTR ldtr.cache.valid = 0;
  ss.cache.valid = 0;
  cs.cache.valid = 0;
  BX_CPU_THIS_PTR ds.cache.valid = 0;
  BX_CPU_THIS_PTR es.cache.valid = 0;
  BX_CPU_THIS_PTR fs.cache.valid = 0;
  BX_CPU_THIS_PTR gs.cache.valid = 0;

  /* if nesting tasks, set the Nested Task flag in the new TSS to 1 */
  if (source == BX_TASK_FROM_CALL_OR_INT) {
    /* ??? */
    }

  /* set the Task Switched flag of the CPU flag to 1 */
  BX_CPU_THIS_PTR cr0.ts = 1;

  /* LDT from the new TSS must be within GDT table limits else #TS(LDT) */
  parse_selector(raw_ldt_selector, &ldt_selector);
  if (ldt_selector.ti) {
    bx_panic("task_switch286: ldt.ti=1\n");
    exception(BX_TS_EXCEPTION, raw_ldt_selector & 0xfffc, 0);
    return;
    }
  fetch_raw_descriptor(&ldt_selector, &dword1, &dword2, 
    BX_TS_EXCEPTION);
  parse_descriptor(dword1, dword2, &ldt_descriptor);

  /* AR byte from LDT descriptor must specify LDT segment else #TS(LDT) */
  if (ldt_descriptor.valid==0 || ldt_descriptor.type!=2) {
    bx_panic("task_switch286: bad LDT segment\n");
    exception(BX_TS_EXCEPTION, raw_ldt_selector & 0xfffc, 0);
    return;
    }
  
  /* AR byte from LDT descriptor must indicate PRESENT else ???#TS(LDT) */
  if (ldt_descriptor.p==0) {
    bx_panic("task_switch286: LDT.p==0\n");
    exception(BX_TS_EXCEPTION, raw_ldt_selector & 0xfffc, 0);
    return;
    }

  /* load LDT cache with new LDT descriptor and set valid bit */
  BX_CPU_THIS_PTR ldtr.selector = ldt_selector;
  BX_CPU_THIS_PTR ldtr.cache    = ldt_descriptor;
  BX_CPU_THIS_PTR ldtr.cache.valid = 1;

  /* set CPL to the RPL of the CS selector in the new TSS */
  cs.selector.rpl = raw_cs_selector & 0x03;

  /* if new stack selector is null #TS(SS) */
  if ( (raw_ss_selector & 0xfffc)==0 ) {
    bx_panic("task_switch286: SS NULL\n");
    exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }

  /* SS selector must be within its descriptor table limits else #TS(SS) */
  parse_selector(raw_ss_selector, &ss_selector);
  fetch_raw_descriptor(&ss_selector, &dword1, &dword2, 
    BX_TS_EXCEPTION);
  parse_descriptor(dword1, dword2, &ss_descriptor);

  /* SS selector RPL must be equal to CPL else #TS(SS) */
  if (ss_selector.rpl != CPL) { /* assigned above */
    bx_panic("task_switch286: SS.rpl != CPL\n");
    exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }

  /* DPL of SS descriptor must equal CPL else #TS(SS) */
  if (ss_descriptor.dpl != CPL) { /* assigned above */
    bx_panic("task_switch286: SS.dpl != CPL\n");
    exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }

  /* SS descriptor AR byte must must indicate writable data segment,
   * else #TS(SS) */
  if (ss_descriptor.valid==0 ||
      ss_descriptor.segment==0 ||
      ss_descriptor.u.segment.executable ||
      ss_descriptor.u.segment.r_w==0) {
    bx_panic("task_switch286: SS not valid\n");
    exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }

  /* SS descriptor AR byte must indicate PRESENT else ??? #SS(SS) */
  if (ss_descriptor.p==0) {
    bx_panic("task_switch286: SS not present\n");
    exception(BX_SS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }

  /* load SS cache with new stack segment and set valid bit */
  ss.selector = ss_selector;
  ss.cache    = ss_descriptor;
  ss.cache.valid = 1;

  /* New CS selector must not be null else #TS(CS) */
  if ( (raw_cs_selector & 0xfffc) == 0 ) {
    bx_panic("task_switch286: CS null\n");
    exception(BX_TS_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }

  /* CS selector must be within its descriptor table limits #TS(CS) */
  parse_selector(raw_cs_selector, &cs_selector);
  fetch_raw_descriptor(&cs_selector, &dword1, &dword2, 
    BX_TS_EXCEPTION);
  parse_descriptor(dword1, dword2, &cs_descriptor);
  
  /* CS descriptor AR byte must indicate code segment else #TS(CS) */
  if (cs_descriptor.valid==0 || cs_descriptor.segment==0 ||
      cs_descriptor.u.segment.executable==0) {
    bx_panic("task_switch286: CS not valid executable seg\n");
    exception(BX_TS_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }

  /* if non-conforming then DPL must equal CPL else #TS(CS) */
  if (cs_descriptor.u.segment.c_ed==0 &&
      cs_descriptor.dpl!=CPL) {
    bx_panic("task_switch286: CS.dpl!=CPL\n");
    exception(BX_TS_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }

  /* if conforming then DPL must be <= CPL else #TS(CS) */
  if (cs_descriptor.u.segment.c_ed &&
      cs_descriptor.dpl>CPL) {
    bx_panic("task_switch286: CS.dpl>CPL\n");
    exception(BX_TS_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }

  /* CS descriptor AR byte must indicate PRESENT else ???#NP(CS) */
  if (cs_descriptor.p==0) {
    bx_panic("task_switch286: CS.p==0\n");
    exception(BX_NP_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }

  /* load CS cache wih new code segment descriptor and set valid bit */
  /* use load_cs() here and above ??? */
  cs.selector = cs_selector;
  cs.cache    = cs_descriptor;
  cs.cache.valid = 1;

  /* for DS, ES: (FS, GS are invalidated above):
   *   if new selector is not null then perform following checks:
   *     index must be within its descriptor table limits else #TS(selector)
   *     AR byte must indicate data or readable code else #TS(selector)
   *     if data or non-conforming code then:
   *       DPL must be >= CPL else #TS(selector)
   *       DPL must be >= RPL else #TS(selector)
   *     AR byte must indicate PRESENT else #NP(selector)
   *     load cache with new segment descriptor and set valid bit
   */

  if ( (raw_ds_selector & 0xfffc) != 0 ) { /* not NULL */
    parse_selector(raw_ds_selector, &ds_selector);
    fetch_raw_descriptor(&ds_selector, &dword1, &dword2, 
      BX_TS_EXCEPTION);
    parse_descriptor(dword1, dword2, &ds_descriptor);
    if (ds_descriptor.valid==0 || ds_descriptor.segment ||
        (ds_descriptor.u.segment.executable &&
         ds_descriptor.u.segment.r_w==0)) {
      bx_panic("task_switch286: DS not valid\n");
      exception(BX_TS_EXCEPTION, raw_ds_selector & 0xfffc, 0);
      return;
      }
    /* if data or non-conforming code */
    if (ds_descriptor.type<12 &&
        (ds_descriptor.dpl<CPL || ds_descriptor.dpl<ds_selector.rpl)) {
      bx_panic("task_switch286: DS.dpl not valid\n");
      exception(BX_TS_EXCEPTION, raw_ds_selector & 0xfffc, 0);
      return;
      }
    if (ds_descriptor.p==0) {
      bx_panic("task_switch286: DS.p==0\n");
      exception(BX_NP_EXCEPTION, raw_ds_selector & 0xfffc, 0);
      return;
      }
    BX_CPU_THIS_PTR ds.selector = ds_selector;
    BX_CPU_THIS_PTR ds.cache    = ds_descriptor;
    }


  if ( (raw_es_selector & 0xfffc) != 0 ) { /* not NULL */
    parse_selector(raw_es_selector, &es_selector);
    fetch_raw_descriptor(&es_selector, &dword1, &dword2, 
      BX_TS_EXCEPTION);
    parse_descriptor(dword1, dword2, &es_descriptor);
    if (es_descriptor.valid==0 || es_descriptor.segment ||
        (es_descriptor.u.segment.executable &&
         es_descriptor.u.segment.r_w==0)) {
      bx_panic("task_switch286: ES not valid\n");
      exception(BX_TS_EXCEPTION, raw_es_selector & 0xfffc, 0);
      return;
      }
    /* if data or non-conforming code */
    if (es_descriptor.type<12 &&
        (es_descriptor.dpl<CPL || es_descriptor.dpl<es_selector.rpl)) {
      bx_panic("task_switch286: ES.dpl not valid\n");
      exception(BX_TS_EXCEPTION, raw_es_selector & 0xfffc, 0);
      return;
      }
    if (es_descriptor.p==0) {
      bx_panic("task_switch286: ES.p==0\n");
      exception(BX_NP_EXCEPTION, raw_es_selector & 0xfffc, 0);
      return;
      }
    BX_CPU_THIS_PTR es.selector = es_selector;
    BX_CPU_THIS_PTR es.cache    = es_descriptor;
    BX_CPU_THIS_PTR es.cache.valid = 1;
    }



  bx_printf("task_switch286(): not complete\n");
  /* check out paging considerations.  Is the whole TSS supposed to
   * be in 1 page?  If that's the case, then only 1 page fault should
   * occur in the beginning? */
}
#endif

#if BX_CPU >= 3
  void
bx_cpu_c::task_switch32(bx_selector_t *selector,
                 bx_descriptor_t *descriptor, unsigned source,
                 Bit32u dword1, Bit32u dword2)
{
  if (v8086_mode()) bx_panic("ctrl_xfer_pro: v8086 mode unsupported\n");

  bx_panic("task_switch386(): not complete\n");
}
#endif




#else  // BX_SUPPORT_TASKING


// for non-support of hardware tasking

#if BX_CPU >= 2
  /* corresponds to SWITCH_TASKS algorithm in Intel documentation */
  void
bx_cpu_c::task_switch16(bx_selector_t *selector,
                 bx_descriptor_t *descriptor, unsigned source,
                 Bit32u dword1, Bit32u dword2)
{
  UNUSED(selector);
  UNUSED(descriptor);
  UNUSED(source);
  UNUSED(dword1);
  UNUSED(dword2);

  bx_printf("task_switch16(): not complete\n");
}
#endif

#if BX_CPU >= 3
  void
bx_cpu_c::task_switch32(bx_selector_t *selector,
                 bx_descriptor_t *descriptor, unsigned source,
                 Bit32u dword1, Bit32u dword2)
{
  UNUSED(selector);
  UNUSED(descriptor);
  UNUSED(source);
  UNUSED(dword1);
  UNUSED(dword2);

  bx_panic("task_switch32(): not complete\n");
}
#endif

#endif // BX_SUPPORT_TASKING
