/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.

BOCHS is commercial software.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/



#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <bochs.h>



// declare these in your software.
bx_cpu_c       bx_cpu;
bx_memory_c    bx_mem;


// these are just for sake of testing, you don't need them.
static Bit8u a20_gate = 1;
static Boolean failure = 0;
static unsigned nmegs = 8;


void bx_vga_mem_write(Bit32u addr, Bit8u value);
void vga_update(void);

  int
main(int argc, char *argv[])
{
  Bit8u value;
  Bit32u i;
  Bit32u value32;
  Bit16u value16;

  bx_bochs_init(argc, argv);

  fprintf(stderr, "=========================================================\n");
  fprintf(stderr, "TEST1: video\n");
  for (i=0; i<2000; i++) {
    value = 'a';
    bx_vga_mem_write(0xb0000 + i, value);
    value = 0x12;
    bx_vga_mem_write(0xb0001 + i, value);
    }

vga_update();
#if 0
  /* Give enough ticks to set off the VGA stuff. */
  for (i=0; i<= 310000; i++) {
    bx_pc_system.TICK();
    }
#endif
  

  fprintf(stderr, "=========================================================\n");
  fprintf(stderr, "TEST2: inp()/outp() calls and the PIC\n");
  fprintf(stderr, "\nlooking at master PIC IMR\n");
  value = bx_mem.inp(0x0021, 1);
  fprintf(stderr, "INP(21) = %02x\n", (unsigned) value);

  fprintf(stderr, "\nallowing IMR IRQ 7\n");
  value &= 0x7f;
  bx_mem.outp(0x0021, value, 1);

  fprintf(stderr, "\nlooking at master PIC IMR\n");
  value = bx_mem.inp(0x0021, 1);
  fprintf(stderr, "INP(21) = %02x\n", (unsigned) value);
  if (value != 0x7f)
    failure = 1;

  /* Give enough ticks to set off the PIT timer. */
  for (i=0; i<= 60000; i++) {
    bx_pc_system.TICK();
    }

  fprintf(stderr, "=========================================================\n");
  fprintf(stderr, "TEST3:\n");
  bx_mem.outp(0x0064, 0xd0, 1); /* read output port */
  value = bx_mem.inp(0x0060, 1);
  fprintf(stderr, "A20 GATE initially of value %u\n",
    (unsigned) (value & 0x02) > 0);
  if ( ((value & 0x02) > 0) != a20_gate ) {
    fprintf(stderr, "\nERROR in A20 logic!\n");
    exit(-1);
    } 

  fprintf(stderr, "\nusing keyboard to disable A20 GATE\n");
  bx_mem.outp(0x0064, 0xd1, 1); /* write output port */
  bx_mem.outp(0x0060, value & 0xfd, 1); /* mask out A20 */
  fprintf(stderr, "A20 GATE now of value %u\n", (unsigned) a20_gate);
  if (a20_gate) failure = 1;

  fprintf(stderr, "\nusing keyboard to enable A20 GATE\n");
  bx_mem.outp(0x0064, 0xd1, 1); /* write output port */
  bx_mem.outp(0x0060, value | 0x02, 1); /* enable A20 */
  fprintf(stderr, "A20 GATE now of value %u\n", (unsigned) a20_gate);
  if (a20_gate==0) failure = 1;
  fprintf(stderr, "=========================================================\n");
  fprintf(stderr, "TEST3:\n");

  fprintf(stderr, "\nenabling IRQ 0\n");
  value = bx_mem.inp(0x0021, 1);
  value &= 0xfe;
  bx_mem.outp(0x0021, value, 1);
  
  fprintf(stderr, "external_set_INTR() should have been called\n");
  fprintf(stderr, "acknowledging interrupt\n");
  value = bx_IAC();
  fprintf(stderr, "CPU would call Int%02xh\n", (unsigned) value);
  if (value != 0x08) failure = 1;

  fprintf(stderr, "=========================================================\n");
  fprintf(stderr, "TEST4:\n");
  fprintf(stderr, "\nwriting 01234567h to 32bit IO port fffch\n");
  bx_mem.outp(0xfffc, 0x01234567, 4);
  value32 = bx_mem.inp(0xfffc, 4);
  fprintf(stderr, "32 bit IO read of port fffch is %08x\n",
    (unsigned) value32);
  if (value32 != 0x01234567) failure = 1;

  fprintf(stderr, "\nwriting abcdh to 16bit IO port fffeh\n");
  bx_mem.outp(0xfffe, 0xabcd, 2);
  value16 = bx_mem.inp(0xfffe, 2);
  fprintf(stderr, "16 bit IO read of port fffeh is %04x\n",
    (unsigned) value16);
  if (value16 != 0xabcd) failure = 1;

  fprintf(stderr, "\nreading 32bit IO port fffc again; shouldn't have"
                  " changed since last 32bit write\n");
  value32 = bx_mem.inp(0xfffc, 4);
  fprintf(stderr, "32 bit IO read of port fffch is %08x\n",
    (unsigned) value32);
  if (value32 != 0x01234567) failure = 1;

  fprintf(stderr, "=========================================================\n");

fprintf(stderr, "\ntriggering IRQ 13\n");
bx_trigger_irq(13);

  fprintf(stderr, "TEST5:\n");
  fprintf(stderr, "\nsending keyboard CPU reset command\n");
  bx_mem.outp(0x0064, 0xfe, 1);

  /* shouldn't get here, since external_reset_cpu() should bomb */
  fprintf(stderr, "ERROR, shouldn't get here!\n");
  return(0);
}

// this used to be bx_set_enable_a20()
  void
bx_memory_c::set_enable_a20(Bit8u value)
{
  fprintf(stderr, ">> external_set_enable_a20() called\n");
  a20_gate = value;
}

// this used to be bx_get_enable_a20()
  Boolean
bx_memory_c::get_enable_a20(void)
{
  fprintf(stderr, ">> external_get_enable_a20() called\n");
  return(a20_gate);
}

// this one is new.  return the number of Kilobytes of memory
// you are emulating.  For instance, for 8Megs return 8192.
  Bit32u
bx_memory_c::get_memory_in_k(void)
{
  return(nmegs * 1024);
}


// this used to be bx_reset_cpu()
  void
bx_cpu_c::reset_cpu(void)
{
  fprintf(stderr, ">> reset_cpu(): keyboard requested RESET!\n");
  fprintf(stderr, ">> hit [return] to end\n");
  getchar();
  if (failure==0)
    fprintf(stderr, "All tests passed!  Could you read the text in the window?\n");
  else
    fprintf(stderr, "One or more tests failed!\n");
  exit(0);
}

// this used to be bx_set_INTR()
  void
bx_cpu_c::set_INTR(Boolean value)
{
  fprintf(stderr, ">> external_set_INTR(%u) called\n", (unsigned) value);
}
