// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com




extern "C" {
#include <time.h>
}

#include "bochs.h"


#if BX_USE_CMOS_SMF
bx_cmos_c bx_cmos;
#define this NULL
#endif


bx_cmos_c::bx_cmos_c(void)
{
  BX_CMOS_THIS s.cmos_mem_address = 0;
 
  BX_CMOS_THIS s.cmos_seconds_alarm = 0xff;
  BX_CMOS_THIS s.cmos_minutes_alarm = 0xff;
  BX_CMOS_THIS s.cmos_hours_alarm = 0xff;
 
  BX_CMOS_THIS s.cmos_shutdown_status = 0x00; /* proceed with normal POST */
  BX_CMOS_THIS s.diagnostic_status = 0x00;
}
 
bx_cmos_c::~bx_cmos_c(void)
{
  // nothing for now
}


  void
bx_cmos_c::init(bx_devices_c *d)
{
  unsigned i;

  /* CMOS RAM & RTC */

  BX_CMOS_THIS devices = d;

  BX_CMOS_THIS devices->register_io_read_handler(this,
       read_handler, 0x0070,
       "CMOS RAM");
  BX_CMOS_THIS devices->register_io_read_handler(this,
       read_handler,
       0x0071,
       "CMOS RAM");
  BX_CMOS_THIS devices->register_io_write_handler(this,
       write_handler,
       0x0070, "CMOS RAM");
  BX_CMOS_THIS devices->register_io_write_handler(this,
       write_handler,
       0x0071, "CMOS RAM");

  for (i=0; i<BX_NUM_CMOS_REGS; i++) {
    BX_CMOS_THIS s.reg[i] = 0;
    }
}


  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_cmos_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_CMOS_SMF
  bx_cmos_c *class_ptr = (bx_cmos_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}

  Bit32u
bx_cmos_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif
  time_t timep;
  struct tm *time_calendar;
  Bit8u val_bcd;
  unsigned year, month, day;

  if (io_len > 1)
    bx_panic("cmos: io read from address %08x len=%u\n",
             (unsigned) address, (unsigned) io_len);

  if (bx_dbg.cmos)
    bx_printf("CMOS read of CMOS register 0x%x\n",
      (unsigned) BX_CMOS_THIS s.cmos_mem_address);


  /* CMOS RAM */
  switch (address) {
    case 0x0071:
      switch (BX_CMOS_THIS s.cmos_mem_address) {
        case 0x01: return(BX_CMOS_THIS s.cmos_seconds_alarm);
        case 0x03: return(BX_CMOS_THIS s.cmos_minutes_alarm);

        case 0x00: /* seconds in BCD format */
          timep = time(NULL);
          time_calendar = localtime(&timep);
          val_bcd = 
            ((time_calendar->tm_sec  / 10) << 4) |
            (time_calendar->tm_sec % 10);
          return(val_bcd);
          break;
        case 0x02: /* minutes in BCD format */
          timep = time(NULL);
          time_calendar = localtime(&timep);
          val_bcd = 
            ((time_calendar->tm_min  / 10) << 4) |
            (time_calendar->tm_min % 10);
          return(val_bcd);
          break;
        case 0x04: /* hours in BCD */
          timep = time(NULL);
          time_calendar = localtime(&timep);
          val_bcd = 
            ((time_calendar->tm_hour  / 10) << 4) |
            (time_calendar->tm_hour % 10);
          return(val_bcd);
          break;

        case 0x05: return(BX_CMOS_THIS s.cmos_hours_alarm);

        case 0x06: // day of the week
          timep = time(NULL);
          time_calendar = localtime(&timep);
          day     = time_calendar->tm_wday + 1; // 0..6 to 1..7
          return(((day / 10) << 4) | (day % 10));

        case 0x07: // day of the month in BCD
          timep = time(NULL);
          time_calendar = localtime(&timep);
          day     = time_calendar->tm_mday;
          return(((day / 10) << 4) | (day % 10));

        case 0x08: // month in BCD
          timep = time(NULL);
          time_calendar = localtime(&timep);
          month   = time_calendar->tm_mon + 1;
          return(((month / 10) << 4) | (month % 10));

        case 0x09: // year in BCD
          timep = time(NULL);
          time_calendar = localtime(&timep);
          year    = time_calendar->tm_year % 100;
          return(((year  / 10) << 4) | (year % 10));

        case 0x0a: return(0x26);
        case 0x0b: // Status Register B
          timep = time(NULL);
          time_calendar = localtime(&timep);
          return(0x02 | (time_calendar->tm_isdst > 0));
        case 0x0c: return(0x00);
        case 0x0d: /* bit7: RTC power stable since last read */
          return(0x80);
        case 0x0e: /* diagnostic status, during POST */
          return(BX_CMOS_THIS s.diagnostic_status); /* no errors */
        case 0x0f: return(BX_CMOS_THIS s.cmos_shutdown_status);

        default:
          if (bx_dbg.cmos) {
            bx_printf("cmos: read cmos[%02xh] = %02x\n",
                      (unsigned) BX_CMOS_THIS s.cmos_mem_address,
                      (unsigned) BX_CMOS_THIS s.reg[BX_CMOS_THIS s.cmos_mem_address]);
            }
          if (BX_CMOS_THIS s.cmos_mem_address>=0x10 && BX_CMOS_THIS s.cmos_mem_address<BX_NUM_CMOS_REGS) {
            return(BX_CMOS_THIS s.reg[BX_CMOS_THIS s.cmos_mem_address]);
            }
          bx_panic("unsupported cmos io read, register(%xh)!\n",
            (unsigned) BX_CMOS_THIS s.cmos_mem_address);
        }
      break;
    default:
      bx_panic("unsupported cmos read, address=%0x%x!\n",
        (unsigned) address);
      break;
    }
  return(0);
}


  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_cmos_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_CMOS_SMF
  bx_cmos_c *class_ptr = (bx_cmos_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_cmos_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_CMOS_SMF

  if (io_len > 1)
    bx_panic("cmos: io write to address %08x len=%u\n",
             (unsigned) address, (unsigned) io_len);

  if (bx_dbg.cmos)
    bx_printf("CMOS write to address: 0x%x = 0x%x\n",
      (unsigned) address, (unsigned) value);


  /* CMOS RAM */
  switch (address) {
    case 0x0070:
      BX_CMOS_THIS s.cmos_mem_address = value & 0x3F;
      break;
    case 0x0071:
      switch (BX_CMOS_THIS s.cmos_mem_address) {
        case 0x01: /* seconds alarm */
          BX_CMOS_THIS s.cmos_seconds_alarm = value;
bx_printf("setting cmos seconds alarm to %x\n", (unsigned) value);
          break;
        case 0x03: /* minutes alarm */
          BX_CMOS_THIS s.cmos_minutes_alarm = value;
bx_printf("setting cmos minutes alarm to %x\n", (unsigned) value);
          break;
        case 0x05: /* hours alarm */
          BX_CMOS_THIS s.cmos_hours_alarm = value;
bx_printf("setting cmos hours alarm to %x\n", (unsigned) value);
          break;
        case 0x0a: /* status register A */
          if (value != 0x26)
            bx_panic("unsupported cmos io write (0x71), case 0x0a = %02x!\n",
              (unsigned) value);
          break;
        case 0x0b: /* status register B */
          if (value!=0x02 && value!=0x03)
            bx_panic("unsupported cmos io write (0x71), case 0x0b = %02x!\n",
              (unsigned) value);
          break;

        case 0x0e: /* diagnostic status */
          BX_CMOS_THIS s.diagnostic_status = value;
          bx_printf("CMOS: write register 0Eh: %02x\n", (unsigned) value);
          break;

	case 0x0f: /* shutdown status */
          switch (value) {
            case 0: /* proceed with normal POST (soft reset) */
              if (bx_dbg.reset)
                bx_printf("CMOS: Reg 0F set to 0: shutdown action = normal POST\n");
              BX_CMOS_THIS s.cmos_shutdown_status = value;
              break;
            case 0x02: /* shutdown after memory test */
              if (bx_dbg.reset)
                bx_printf("CMOS: Reg 0Fh: request to change shutdown action"
                             " to shutdown after memory test\n");
              BX_CMOS_THIS s.cmos_shutdown_status = value;
              break;
            case 4: /* jump to disk bootstrap routine */
              bx_printf("CMOS: Reg 0Fh: request to change shutdown action "
                             "to jump to disk bootstrap routine.\n");
              BX_CMOS_THIS s.cmos_shutdown_status = value;
              break;
            case 9: /* return to BIOS extended memory block move
                       (interrupt 15h, func 87h was in progress) */
              if (bx_dbg.reset)
                bx_printf("CMOS: Reg 0Fh: request to change shutdown action "
                             "to return to BIOS extended memory block move.\n");
              BX_CMOS_THIS s.cmos_shutdown_status = value;
              break;
            case 0x0a: /* jump to DWORD pointer at 40:67 */
              if (bx_dbg.reset)
                bx_printf("CMOS: Reg 0Fh: request to change shutdown action"
                             " to jump to DWORD at 40:67\n");
              BX_CMOS_THIS s.cmos_shutdown_status = value;
              break;
            default:
              bx_panic("unsupported cmos io write to reg F, case %x!\n",
                (unsigned) value);
              break;
            }
          break;

        default:
          if (BX_CMOS_THIS s.cmos_mem_address>=0x10 && BX_CMOS_THIS s.cmos_mem_address<BX_NUM_CMOS_REGS) {
            bx_printf("CMOS: write reg %02xh: value = %02xh\n",
              (unsigned) BX_CMOS_THIS s.cmos_mem_address, (unsigned) value);
            BX_CMOS_THIS s.reg[BX_CMOS_THIS s.cmos_mem_address] = value;
            break;
            }
          bx_printf("unsupported cmos io write, register(%xh)=%02x!\n",
            (unsigned) BX_CMOS_THIS s.cmos_mem_address, (unsigned) value);
        }
      break;
    }
}


  void
bx_cmos_c::checksum_cmos(void)
{
  unsigned i;
  Bit16u sum;

  sum = 0;
  for (i=0x10; i<=0x2d; i++) {
    sum += BX_CMOS_THIS s.reg[i];
    }
  BX_CMOS_THIS s.reg[0x2e] = (sum >> 8) & 0xff; /* checksum high */
  BX_CMOS_THIS s.reg[0x2f] = (sum & 0xff);      /* checksum low */

bx_printf("cmos: cmos[0x14] = %02x\n", (unsigned) BX_CMOS_THIS s.reg[0x14]);
}
